# Copyright 2018-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["install_mask_dir", "InstallMask"]

import collections
import errno
import fnmatch
import operator

from portage import os, _unicode_decode
from portage.exception import (
    FileNotFound,
    IsADirectory,
    OperationNotPermitted,
    PermissionDenied,
    ReadOnlyFileSystem,
)
from portage.util import normalize_path


def _defaultdict_tree():
    return collections.defaultdict(_defaultdict_tree)


_pattern = collections.namedtuple(
    "_pattern",
    (
        "orig_index",
        "is_inclusive",
        "pattern",
        "leading_slash",
    ),
)


class InstallMask:
    def __init__(self, install_mask):
        """
        @param install_mask: INSTALL_MASK value
        @type install_mask: str
        """
        # Patterns not anchored with leading slash
        self._unanchored = []

        # Patterns anchored with leading slash are indexed by leading
        # non-glob components, making it possible to minimize the
        # number of fnmatch calls. For example:
        # /foo*/bar -> {'.': ['/foo*/bar']}
        # /foo/bar* -> {'foo': {'.': ['/foo/bar*']}}
        # /foo/bar/ -> {'foo': {'bar': {'.': ['/foo/bar/']}}}
        self._anchored = _defaultdict_tree()
        for orig_index, pattern in enumerate(install_mask.split()):
            # if pattern starts with -, possibly exclude this path
            is_inclusive = not pattern.startswith("-")
            if not is_inclusive:
                pattern = pattern[1:]
            pattern_obj = _pattern(
                orig_index, is_inclusive, pattern, pattern.startswith("/")
            )
            # absolute path pattern
            if pattern_obj.leading_slash:
                current_dir = self._anchored
                for component in list(filter(None, pattern.split("/"))):
                    if "*" in component:
                        break
                    else:
                        current_dir = current_dir[component]
                current_dir.setdefault(".", []).append(pattern_obj)

            # filename
            else:
                self._unanchored.append(pattern_obj)

    def _iter_relevant_patterns(self, path):
        """
        Iterate over patterns that may be relevant for the given path.

        Patterns anchored with leading / are indexed by leading
        non-glob components, making it possible to minimize the
        number of fnmatch calls.
        """
        current_dir = self._anchored
        components = list(filter(None, path.split("/")))
        patterns = []
        patterns.extend(current_dir.get(".", []))
        for component in components:
            next_dir = current_dir.get(component, None)
            if next_dir is None:
                break
            current_dir = next_dir
            patterns.extend(current_dir.get(".", []))

        if patterns:
            # Sort by original pattern index, since order matters for
            # non-inclusive patterns.
            patterns.extend(self._unanchored)
            if any(not pattern.is_inclusive for pattern in patterns):
                patterns.sort(key=operator.attrgetter("orig_index"))
            return iter(patterns)

        return iter(self._unanchored)

    def match(self, path):
        """
        @param path: file path relative to ${ED}
        @type path: str
        @rtype: bool
        @return: True if path matches INSTALL_MASK, False otherwise
        """
        ret = False

        for pattern_obj in self._iter_relevant_patterns(path):
            is_inclusive, pattern = pattern_obj.is_inclusive, pattern_obj.pattern
            # absolute path pattern
            if pattern_obj.leading_slash:
                # handle trailing slash for explicit directory match
                if path.endswith("/"):
                    pattern = pattern.rstrip("/") + "/"
                # match either exact path or one of parent dirs
                # the latter is done via matching pattern/*
                if fnmatch.fnmatch(path, pattern[1:]) or fnmatch.fnmatch(
                    path, pattern[1:].rstrip("/") + "/*"
                ):
                    ret = is_inclusive
            # filename
            else:
                if fnmatch.fnmatch(os.path.basename(path), pattern):
                    ret = is_inclusive
        return ret


_exc_map = {
    errno.EISDIR: IsADirectory,
    errno.ENOENT: FileNotFound,
    errno.EPERM: OperationNotPermitted,
    errno.EACCES: PermissionDenied,
    errno.EROFS: ReadOnlyFileSystem,
}


def _raise_exc(e):
    """
    Wrap OSError with portage.exception wrapper exceptions, with
    __cause__ chaining when python supports it.

    @param e: os exception
    @type e: OSError
    @raise PortageException: portage.exception wrapper exception
    """
    wrapper_cls = _exc_map.get(e.errno)
    if wrapper_cls is None:
        raise
    wrapper = wrapper_cls(str(e))
    wrapper.__cause__ = e
    raise wrapper


def install_mask_dir(base_dir, install_mask, onerror=None):
    """
    Remove files and directories matched by INSTALL_MASK.

    @param base_dir: directory path corresponding to ${ED}
    @type base_dir: str
    @param install_mask: INSTALL_MASK configuration
    @type install_mask: InstallMask
    """
    onerror = onerror or _raise_exc
    base_dir = normalize_path(base_dir)
    base_dir_len = len(base_dir) + 1
    dir_stack = []

    # Remove masked files.
    todo = [base_dir]
    while todo:
        parent = todo.pop()
        try:
            parent = _unicode_decode(parent, errors="strict")
        except UnicodeDecodeError:
            continue

        dir_stack.append(parent)
        for entry in os.scandir(parent):
            try:
                abs_path = _unicode_decode(entry.path, errors="strict")
            except UnicodeDecodeError:
                continue

            if entry.is_dir(follow_symlinks=False):
                todo.append(entry.path)
            elif install_mask.match(abs_path[base_dir_len:]):
                try:
                    os.unlink(entry.path)
                except OSError as e:
                    onerror(e)

    # Remove masked dirs (unless non-empty due to exclusions).
    while True:
        try:
            dir_path = dir_stack.pop()
        except IndexError:
            break

        if install_mask.match(dir_path[base_dir_len:] + "/"):
            try:
                os.rmdir(dir_path)
            except OSError:
                pass



# Copyright 2014-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import sys
import logging
import grp
import pwd
import warnings

import portage
from portage import os
from portage.progress import ProgressBar

# from portage.emaint.defaults import DEFAULT_OPTIONS
from portage.util import writemsg, writemsg_level
from portage.output import create_color_func

good = create_color_func("GOOD")
bad = create_color_func("BAD")
warn = create_color_func("WARN")
from portage.package.ebuild.doebuild import _check_temp_dir
from portage.metadata import action_metadata
from portage.util.hooks import get_hooks_from_dir
from portage.util._async.AsyncFunction import AsyncFunction
from portage import _unicode_decode
from _emerge.CompositeTask import CompositeTask


class TaskHandler:
    """Handles the running of the tasks it is given"""

    def __init__(self, show_progress_bar=True, verbose=True, callback=None):
        self.show_progress_bar = show_progress_bar
        self.verbose = verbose
        self.callback = callback
        self.isatty = os.environ.get("TERM") != "dumb" and sys.stdout.isatty()
        self.progress_bar = ProgressBar(
            self.isatty, title="Portage-Sync", max_desc_length=27
        )

    def run_tasks(self, tasks, func, status=None, verbose=True, options=None):
        """Runs the module tasks"""
        # Ensure we have a task and function
        assert tasks
        assert func
        for task in tasks:
            inst = task()
            show_progress = self.show_progress_bar and self.isatty
            # check if the function is capable of progressbar
            # and possibly override it off
            if show_progress and hasattr(inst, "can_progressbar"):
                show_progress = inst.can_progressbar(func)
            if show_progress:
                self.progress_bar.reset()
                self.progress_bar.set_label(func + " " + inst.name())
                onProgress = self.progress_bar.start()
            else:
                onProgress = None
            kwargs = {
                "onProgress": onProgress,
                # pass in a copy of the options so a module can not pollute or change
                # them for other tasks if there is more to do.
                "options": options.copy(),
            }
            result = getattr(inst, func)(**kwargs)
            if show_progress:
                # make sure the final progress is displayed
                self.progress_bar.display()
                print()
                self.progress_bar.stop()
            if self.callback:
                self.callback(result)


def print_results(results):
    if results:
        print()
        print("\n".join(results))
        print("\n")


class SyncManager:
    """Main sync control module"""

    def __init__(self, settings, logger):
        self.settings = settings
        self.logger = logger
        # Similar to emerge, sync needs a default umask so that created
        # files have sane permissions.
        os.umask(0o22)

        self.hooks = {}
        for _dir in ["repo.postsync.d", "postsync.d"]:
            hooks = get_hooks_from_dir(_dir, prefix=self.settings["PORTAGE_CONFIGROOT"])
            self.hooks[_dir] = hooks

    @property
    def module_controller(self):
        # Not stored as local attribute because it's not picklable.
        return portage.sync.module_controller

    @property
    def module_names(self):
        return self.module_controller.module_names

    def __getattr__(self, name):
        if name == "async":
            warnings.warn(
                "portage.sync.controller.SyncManager.async "
                "has been renamed to sync_async",
                DeprecationWarning,
                stacklevel=2,
            )
            return self.sync_async
        raise AttributeError(name)

    def get_module_descriptions(self, mod):
        desc = self.module_controller.get_func_descriptions(mod)
        if desc:
            return desc
        return []

    def sync_async(self, emerge_config=None, repo=None, master_hooks=True):
        self.emerge_config = emerge_config
        self.settings, self.trees, self.mtimedb = emerge_config
        self.xterm_titles = "notitles" not in self.settings.features
        self.portdb = self.trees[self.settings["EROOT"]]["porttree"].dbapi
        return SyncRepo(
            sync_task=AsyncFunction(
                target=self.sync,
                kwargs=dict(
                    emerge_config=emerge_config, repo=repo, master_hooks=master_hooks
                ),
            ),
            sync_callback=self._sync_callback,
        )

    def sync(self, emerge_config=None, repo=None, master_hooks=True):
        from portage.sync.revision_history import get_repo_revision_history

        self.callback = None
        self.repo = repo
        self.exitcode = 1
        self.updatecache_flg = False
        hooks_enabled = master_hooks or not repo.sync_hooks_only_on_change
        if repo.sync_type in self.module_names:
            tasks = [self.module_controller.get_class(repo.sync_type)]
        else:
            msg = "\n{}: Sync module '{}' is not an installed/known type'\n".format(
                bad("ERROR"),
                repo.sync_type,
            )
            return self.exitcode, msg, self.updatecache_flg, hooks_enabled

        rval = self.pre_sync(repo)
        if rval != os.EX_OK:
            return rval, None, self.updatecache_flg, hooks_enabled

        # need to pass the kwargs dict to the modules
        # so they are available if needed.
        task_opts = {
            "emerge_config": emerge_config,
            "logger": self.logger,
            "portdb": self.portdb,
            "repo": repo,
            "settings": self.settings,
            "spawn_kwargs": self.spawn_kwargs,
            "usersync_uid": self.usersync_uid,
            "xterm_titles": self.xterm_titles,
        }
        func = "sync"
        status = None
        taskmaster = TaskHandler(callback=self.do_callback)
        taskmaster.run_tasks(tasks, func, status, options=task_opts)
        get_repo_revision_history(self.settings["EROOT"], [repo])

        if master_hooks or self.updatecache_flg or not repo.sync_hooks_only_on_change:
            hooks_enabled = True
            self.perform_post_sync_hook(repo.name, repo.sync_uri, repo.location)

        return self.exitcode, None, self.updatecache_flg, hooks_enabled

    def do_callback(self, result):
        # print("result:", result, "callback()", self.callback)
        exitcode, updatecache_flg = result
        self.exitcode = exitcode
        self.updatecache_flg = updatecache_flg
        if exitcode == 0:
            msg = f"=== Sync completed for {self.repo.name}"
            self.logger(self.xterm_titles, msg)
            writemsg_level(msg + "\n")
        if self.callback:
            self.callback(exitcode, updatecache_flg)

    def perform_post_sync_hook(self, reponame, dosyncuri="", repolocation=""):
        succeeded = os.EX_OK
        if reponame:
            _hooks = self.hooks["repo.postsync.d"]
        else:
            _hooks = self.hooks["postsync.d"]
        for filepath in _hooks:
            writemsg_level(
                f"Spawning post_sync hook: {_unicode_decode(_hooks[filepath])}\n",
                level=logging.ERROR,
                noiselevel=4,
            )
            if reponame:
                retval = portage.process.spawn(
                    [filepath, reponame, dosyncuri, repolocation],
                    env=self.settings.environ(),
                )
            else:
                retval = portage.process.spawn([filepath], env=self.settings.environ())
            if retval != os.EX_OK:
                writemsg_level(
                    " %s Spawn failed for: %s, %s\n"
                    % (bad("*"), _unicode_decode(_hooks[filepath]), filepath),
                    level=logging.ERROR,
                    noiselevel=-1,
                )
                succeeded = retval
        return succeeded

    def pre_sync(self, repo):
        msg = f">>> Syncing repository '{repo.name}' into '{repo.location}'..."
        self.logger(self.xterm_titles, msg)
        writemsg_level(msg + "\n")
        try:
            st = os.stat(repo.location)
        except OSError:
            st = None

        self.usersync_uid = None
        spawn_kwargs = {}
        # Redirect command stderr to stdout, in order to prevent
        # spurious cron job emails (bug 566132).
        spawn_kwargs["fd_pipes"] = {
            0: portage._get_stdin().fileno(),
            1: sys.__stdout__.fileno(),
            2: sys.__stdout__.fileno(),
        }
        spawn_kwargs["env"] = self.settings.environ()
        if repo.sync_user is not None:

            def get_sync_user_data(sync_user):
                user = None
                group = None
                home = None
                logname = None

                spl = sync_user.split(":", 1)
                if spl[0]:
                    username = spl[0]
                    try:
                        try:
                            pw = pwd.getpwnam(username)
                        except KeyError:
                            pw = pwd.getpwuid(int(username))
                    except (ValueError, KeyError):
                        writemsg(
                            f"!!! User '{username}' invalid or does not exist\n",
                            noiselevel=-1,
                        )
                        return (logname, user, group, home)
                    user = pw.pw_uid
                    group = pw.pw_gid
                    home = pw.pw_dir
                    logname = pw.pw_name

                if len(spl) > 1:
                    groupname = spl[1]
                    try:
                        try:
                            gp = grp.getgrnam(groupname)
                        except KeyError:
                            gp = grp.getgrgid(int(groupname))
                    except (ValueError, KeyError):
                        writemsg(
                            f"!!! Group '{groupname}' invalid or does not exist\n",
                            noiselevel=-1,
                        )
                        return (logname, user, group, home)

                    group = gp.gr_gid

                return (logname, user, group, home)

            # user or user:group
            (logname, uid, gid, home) = get_sync_user_data(repo.sync_user)
            if uid is not None:
                spawn_kwargs["uid"] = uid
                self.usersync_uid = uid
            if gid is not None:
                spawn_kwargs["gid"] = gid
                spawn_kwargs["groups"] = [gid]
            if home is not None:
                spawn_kwargs["env"]["HOME"] = home
            if logname is not None:
                spawn_kwargs["env"]["LOGNAME"] = logname

        if st is None:
            perms = {"mode": 0o755}
            # respect sync-user if set
            if "umask" in spawn_kwargs:
                perms["mode"] &= ~spawn_kwargs["umask"]
            if "uid" in spawn_kwargs:
                perms["uid"] = spawn_kwargs["uid"]
            if "gid" in spawn_kwargs:
                perms["gid"] = spawn_kwargs["gid"]

            portage.util.ensure_dirs(repo.location, **perms)
            st = os.stat(repo.location)

        if (
            repo.sync_user is None
            and "usersync" in self.settings.features
            and portage.data.secpass >= 2
            and (
                st.st_uid != os.getuid()
                and st.st_mode & 0o700
                or st.st_gid != os.getgid()
                and st.st_mode & 0o070
            )
        ):
            try:
                pw = pwd.getpwuid(st.st_uid)
            except KeyError:
                pass
            else:
                # Drop privileges when syncing, in order to match
                # existing uid/gid settings.
                self.usersync_uid = st.st_uid
                spawn_kwargs["uid"] = st.st_uid
                spawn_kwargs["gid"] = st.st_gid
                spawn_kwargs["groups"] = [st.st_gid]
                spawn_kwargs["env"]["HOME"] = pw.pw_dir
                spawn_kwargs["env"]["LOGNAME"] = pw.pw_name
                umask = 0o002
                if not st.st_mode & 0o020:
                    umask = umask | 0o020
                spawn_kwargs["umask"] = umask
        # override the defaults when sync_umask is set
        if repo.sync_umask is not None:
            spawn_kwargs["umask"] = int(repo.sync_umask, 8)
        spawn_kwargs.setdefault("umask", 0o022)
        self.spawn_kwargs = spawn_kwargs

        if self.usersync_uid is not None:
            # PORTAGE_TMPDIR is used below, so validate it and
            # bail out if necessary.
            rval = _check_temp_dir(self.settings)
            if rval != os.EX_OK:
                return rval

        os.umask(0o022)
        return os.EX_OK

    def _sync_callback(self, proc):
        """
        This is called in the parent process, serially, for each of the
        sync jobs when they complete. Some cache backends such as sqlite
        may require that cache access be performed serially in the
        parent process like this.
        """
        repo = proc.kwargs["repo"]
        exitcode = proc.returncode
        updatecache_flg = False
        if proc.returncode == os.EX_OK:
            exitcode, message, updatecache_flg, hooks_enabled = proc.result

        if updatecache_flg and "metadata-transfer" not in self.settings.features:
            updatecache_flg = False

        if updatecache_flg and os.path.exists(
            os.path.join(repo.location, "metadata", "md5-cache")
        ):
            # Only update cache for repo.location since that's
            # the only one that's been synced here.
            action_metadata(
                self.settings,
                self.portdb,
                self.emerge_config.opts,
                porttrees=[repo.location],
            )


class SyncRepo(CompositeTask):
    """
    Encapsulates a sync operation and the callback which executes afterwards,
    so both can be considered as a single composite task. This is useful
    since we don't want to consider a particular repo's sync operation as
    complete until after the callback has executed (bug 562264).

    The kwargs and result properties expose attributes that are accessed
    by SyncScheduler.
    """

    __slots__ = ("sync_task", "sync_callback")

    @property
    def kwargs(self):
        return self.sync_task.kwargs

    @property
    def result(self):
        return self.sync_task.result

    def _start(self):
        self._start_task(self.sync_task, self._sync_task_exit)

    def _sync_task_exit(self, sync_task):
        self._current_task = None
        self.returncode = sync_task.returncode
        self.sync_callback(self.sync_task)
        self._async_wait()



# Copyright 2016 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2
#
# This module provides an extended subset of the asyncio.futures.Futures
# interface.

__all__ = (
    "CancelledError",
    "ExtendedFuture",
    "InvalidStateError",
)

import concurrent.futures
from concurrent.futures import Future, CancelledError

try:
    from concurrent.futures import InvalidStateError
except ImportError:
    from portage.util.futures.futures import InvalidStateError

# Create our one time settable unset constant
UNSET_CONST = Future()
UNSET_CONST.set_result(object())


class ExtendedFuture(Future):
    """Extended Future class adding convienince get and set operations with
    default result capabilities for unset result().  It also adds pass
    capability for duplicate set_result() calls.
    """

    def __init__(self, default_result=UNSET_CONST.result()):
        """Class init

        @param default_result: Optional data type/value to return in the event
                               of a result() call when result has not yet been
                               set.
        """
        self.default_result = default_result
        super().__init__()
        self.set = self.set_result

    def set_result(self, data, ignore_InvalidState=False):
        """Set the Future's result to the data, optionally don't raise
        an error for 'InvalidStateError' errors

        @param ignore_exception: Boolean
        """
        if ignore_InvalidState:
            try:
                super().set_result(data)
            except InvalidStateError:
                pass
        else:
            super().set_result(data)

    def get(self, default=UNSET_CONST.result()):
        """Convenience function to wrap result() but adds an optional
        default value to return rather than raise an InvalidStateError

        @param default: Optional override for the classwide default_result
        @returns: the result data or the default value, raisies an exception
                  if result is unset and no default is defined.
        """
        if default is not UNSET_CONST.result():
            pass
        elif self.default_result is not UNSET_CONST.result():
            default = self.default_result
        if default is not UNSET_CONST.result():
            try:
                data = self.result()
            except InvalidStateError:
                data = default
        else:
            data = self.result()
        return data

    def exception(self):
        try:
            return super().exception(timeout=0)
        except concurrent.futures.TimeoutError:
            raise InvalidStateError

    def result(self):
        try:
            return super().result(timeout=0)
        except concurrent.futures.TimeoutError:
            raise InvalidStateError



# SPDX-License-Identifier: Apache-2.0
# Copyright © 2022-2024 Intel Corporation

"""Interpreter for converting Cargo Toml definitions to Meson AST

There are some notable limits here. We don't even try to convert something with
a build.rs: there's so few limits on what Cargo allows a build.rs (basically
none), and no good way for us to convert them. In that case, an actual meson
port will be required.
"""

from __future__ import annotations
import dataclasses
import os
import collections
import urllib.parse
import itertools
import typing as T

from . import builder, version, cfg
from .toml import load_toml, TomlImplementationMissing
from .manifest import Manifest, CargoLock, fixup_meson_varname
from ..mesonlib import MesonException, MachineChoice
from .. import coredata, mlog
from ..wrap.wrap import PackageDefinition

if T.TYPE_CHECKING:
    from . import raw
    from .. import mparser
    from .manifest import Dependency, SystemDependency
    from ..environment import Environment
    from ..interpreterbase import SubProject
    from ..compilers.rust import RustCompiler

def _dependency_name(package_name: str, api: str, suffix: str = '-rs') -> str:
    basename = package_name[:-len(suffix)] if package_name.endswith(suffix) else package_name
    return f'{basename}-{api}{suffix}'


def _dependency_varname(package_name: str) -> str:
    return f'{fixup_meson_varname(package_name)}_dep'


def _extra_args_varname() -> str:
    return 'extra_args'


def _extra_deps_varname() -> str:
    return 'extra_deps'


@dataclasses.dataclass
class PackageState:
    manifest: Manifest
    downloaded: bool = False
    features: T.Set[str] = dataclasses.field(default_factory=set)
    required_deps: T.Set[str] = dataclasses.field(default_factory=set)
    optional_deps_features: T.Dict[str, T.Set[str]] = dataclasses.field(default_factory=lambda: collections.defaultdict(set))


@dataclasses.dataclass(frozen=True)
class PackageKey:
    package_name: str
    api: str


class Interpreter:
    def __init__(self, env: Environment) -> None:
        self.environment = env
        self.host_rustc = T.cast('RustCompiler', self.environment.coredata.compilers[MachineChoice.HOST]['rust'])
        # Map Cargo.toml's subdir to loaded manifest.
        self.manifests: T.Dict[str, Manifest] = {}
        # Map of cargo package (name + api) to its state
        self.packages: T.Dict[PackageKey, PackageState] = {}
        # Rustc's config
        self.cfgs = self._get_cfgs()

    def get_build_def_files(self) -> T.List[str]:
        return [os.path.join(subdir, 'Cargo.toml') for subdir in self.manifests]

    def interpret(self, subdir: str) -> mparser.CodeBlockNode:
        manifest = self._load_manifest(subdir)
        pkg, cached = self._fetch_package(manifest.package.name, manifest.package.api)
        if not cached:
            # This is an entry point, always enable the 'default' feature.
            # FIXME: We should have a Meson option similar to `cargo build --no-default-features`
            self._enable_feature(pkg, 'default')

        # Build an AST for this package
        filename = os.path.join(self.environment.source_dir, subdir, 'Cargo.toml')
        build = builder.Builder(filename)
        ast = self._create_project(pkg, build)
        ast += [
            build.assign(build.function('import', [build.string('rust')]), 'rust'),
            build.function('message', [
                build.string('Enabled features:'),
                build.array([build.string(f) for f in pkg.features]),
            ]),
        ]
        ast += self._create_dependencies(pkg, build)
        ast += self._create_meson_subdir(build)

        if pkg.manifest.lib:
            for crate_type in pkg.manifest.lib.crate_type:
                ast.extend(self._create_lib(pkg, build, crate_type))

        return build.block(ast)

    def _fetch_package(self, package_name: str, api: str) -> T.Tuple[PackageState, bool]:
        key = PackageKey(package_name, api)
        pkg = self.packages.get(key)
        if pkg:
            return pkg, True
        meson_depname = _dependency_name(package_name, api)
        subdir, _ = self.environment.wrap_resolver.resolve(meson_depname)
        subprojects_dir = os.path.join(subdir, 'subprojects')
        self.environment.wrap_resolver.load_and_merge(subprojects_dir, T.cast('SubProject', meson_depname))
        manifest = self._load_manifest(subdir)
        downloaded = \
            meson_depname in self.environment.wrap_resolver.wraps and \
            self.environment.wrap_resolver.wraps[meson_depname].type is not None
        pkg = PackageState(manifest, downloaded)
        self.packages[key] = pkg
        # Merge target specific dependencies that are enabled
        for condition, dependencies in manifest.target.items():
            if cfg.eval_cfg(condition, self.cfgs):
                manifest.dependencies.update(dependencies)
        # Fetch required dependencies recursively.
        for depname, dep in manifest.dependencies.items():
            if not dep.optional:
                self._add_dependency(pkg, depname)
        return pkg, False

    def _dep_package(self, dep: Dependency) -> PackageState:
        return self.packages[PackageKey(dep.package, dep.api)]

    def _load_manifest(self, subdir: str) -> Manifest:
        manifest_ = self.manifests.get(subdir)
        if not manifest_:
            path = os.path.join(self.environment.source_dir, subdir)
            filename = os.path.join(path, 'Cargo.toml')
            toml = load_toml(filename)
            if 'package' in toml:
                raw_manifest = T.cast('raw.Manifest', toml)
                manifest_ = Manifest.from_raw(raw_manifest, path)
                self.manifests[subdir] = manifest_
            else:
                raise MesonException(f'{subdir}/Cargo.toml does not have [package] section')
        return manifest_

    def _add_dependency(self, pkg: PackageState, depname: str) -> None:
        if depname in pkg.required_deps:
            return
        dep = pkg.manifest.dependencies.get(depname)
        if not dep:
            if depname in itertools.chain(pkg.manifest.dev_dependencies, pkg.manifest.build_dependencies):
                # FIXME: Not supported yet
                return
            raise MesonException(f'Dependency {depname} not defined in {pkg.manifest.package.name} manifest')
        pkg.required_deps.add(depname)
        dep_pkg, _ = self._fetch_package(dep.package, dep.api)
        if dep.default_features:
            self._enable_feature(dep_pkg, 'default')
        for f in dep.features:
            self._enable_feature(dep_pkg, f)
        for f in pkg.optional_deps_features[depname]:
            self._enable_feature(dep_pkg, f)

    def _enable_feature(self, pkg: PackageState, feature: str) -> None:
        if feature in pkg.features:
            return
        pkg.features.add(feature)
        # A feature can also be a dependency.
        if feature in pkg.manifest.dependencies:
            self._add_dependency(pkg, feature)
        # Recurse on extra features and dependencies this feature pulls.
        # https://doc.rust-lang.org/cargo/reference/features.html#the-features-section
        for f in pkg.manifest.features.get(feature, []):
            if '/' in f:
                depname, dep_f = f.split('/', 1)
                if depname[-1] == '?':
                    depname = depname[:-1]
                    if depname in pkg.required_deps:
                        dep = pkg.manifest.dependencies[depname]
                        dep_pkg = self._dep_package(dep)
                        self._enable_feature(dep_pkg, dep_f)
                    else:
                        # This feature will be enabled only if that dependency
                        # is later added.
                        pkg.optional_deps_features[depname].add(dep_f)
                else:
                    self._add_dependency(pkg, depname)
                    dep = pkg.manifest.dependencies.get(depname)
                    if dep:
                        dep_pkg = self._dep_package(dep)
                        self._enable_feature(dep_pkg, dep_f)
            elif f.startswith('dep:'):
                self._add_dependency(pkg, f[4:])
            else:
                self._enable_feature(pkg, f)

    def _get_cfgs(self) -> T.Dict[str, str]:
        cfgs = self.host_rustc.get_cfgs().copy()
        rustflags = self.environment.coredata.get_external_args(MachineChoice.HOST, 'rust')
        rustflags_i = iter(rustflags)
        for i in rustflags_i:
            if i == '--cfg':
                cfgs.append(next(rustflags_i))
        return dict(self._split_cfg(i) for i in cfgs)

    @staticmethod
    def _split_cfg(cfg: str) -> T.Tuple[str, str]:
        pair = cfg.split('=', maxsplit=1)
        value = pair[1] if len(pair) > 1 else ''
        if value and value[0] == '"':
            value = value[1:-1]
        return pair[0], value

    def _create_project(self, pkg: PackageState, build: builder.Builder) -> T.List[mparser.BaseNode]:
        """Create the project() function call

        :param pkg: The package to generate from
        :param build: The AST builder
        :return: a list nodes
        """
        default_options: T.List[mparser.BaseNode] = []
        default_options.append(build.string(f'rust_std={pkg.manifest.package.edition}'))
        default_options.append(build.string(f'build.rust_std={pkg.manifest.package.edition}'))
        if pkg.downloaded:
            default_options.append(build.string('warning_level=0'))

        args: T.List[mparser.BaseNode] = []
        args.extend([
            build.string(pkg.manifest.package.name),
            build.string('rust'),
        ])
        kwargs: T.Dict[str, mparser.BaseNode] = {
            'version': build.string(pkg.manifest.package.version),
            # Always assume that the generated meson is using the latest features
            # This will warn when when we generate deprecated code, which is helpful
            # for the upkeep of the module
            'meson_version': build.string(f'>= {coredata.stable_version}'),
            'default_options': build.array(default_options),
        }
        if pkg.manifest.package.license:
            kwargs['license'] = build.string(pkg.manifest.package.license)
        elif pkg.manifest.package.license_file:
            kwargs['license_files'] = build.string(pkg.manifest.package.license_file)

        return [build.function('project', args, kwargs)]

    def _create_dependencies(self, pkg: PackageState, build: builder.Builder) -> T.List[mparser.BaseNode]:
        ast: T.List[mparser.BaseNode] = []
        for depname in pkg.required_deps:
            dep = pkg.manifest.dependencies[depname]
            ast += self._create_dependency(dep, build)
        ast.append(build.assign(build.array([]), 'system_deps_args'))
        for name, sys_dep in pkg.manifest.system_dependencies.items():
            if sys_dep.enabled(pkg.features):
                ast += self._create_system_dependency(name, sys_dep, build)
        return ast

    def _create_system_dependency(self, name: str, dep: SystemDependency, build: builder.Builder) -> T.List[mparser.BaseNode]:
        # TODO: handle feature_overrides
        kw = {
            'version': build.array([build.string(s) for s in dep.meson_version]),
            'required': build.bool(not dep.optional),
        }
        varname = f'{fixup_meson_varname(name)}_system_dep'
        cfg = f'system_deps_have_{fixup_meson_varname(name)}'
        return [
            build.assign(
                build.function(
                    'dependency',
                    [build.string(dep.name)],
                    kw,
                ),
                varname,
            ),
            build.if_(
                build.method('found', build.identifier(varname)), build.block([
                    build.plusassign(
                        build.array([build.string('--cfg'), build.string(cfg)]),
                        'system_deps_args'
                    ),
                ])
            ),
        ]

    def _create_dependency(self, dep: Dependency, build: builder.Builder) -> T.List[mparser.BaseNode]:
        pkg = self._dep_package(dep)
        kw = {
            'version': build.array([build.string(s) for s in dep.meson_version]),
        }
        # Lookup for this dependency with the features we want in default_options kwarg.
        #
        # However, this subproject could have been previously configured with a
        # different set of features. Cargo collects the set of features globally
        # but Meson can only use features enabled by the first call that triggered
        # the configuration of that subproject.
        #
        # Verify all features that we need are actually enabled for that dependency,
        # otherwise abort with an error message. The user has to set the corresponding
        # option manually with -Dxxx-rs:feature-yyy=true, or the main project can do
        # that in its project(..., default_options: ['xxx-rs:feature-yyy=true']).
        return [
            # xxx_dep = dependency('xxx', version : ...)
            build.assign(
                build.function(
                    'dependency',
                    [build.string(_dependency_name(dep.package, dep.api))],
                    kw,
                ),
                _dependency_varname(dep.package),
            ),
            # actual_features = xxx_dep.get_variable('features', default_value : '').split(',')
            build.assign(
                build.method(
                    'split',
                    build.method(
                        'get_variable',
                        build.identifier(_dependency_varname(dep.package)),
                        [build.string('features')],
                        {'default_value': build.string('')}
                    ),
                    [build.string(',')],
                ),
                'actual_features'
            ),
            # needed_features = [f1, f2, ...]
            # foreach f : needed_features
            #   if f not in actual_features
            #     error()
            #   endif
            # endforeach
            build.assign(build.array([build.string(f) for f in pkg.features]), 'needed_features'),
            build.foreach(['f'], build.identifier('needed_features'), build.block([
                build.if_(build.not_in(build.identifier('f'), build.identifier('actual_features')), build.block([
                    build.function('error', [
                        build.string('Dependency'),
                        build.string(_dependency_name(dep.package, dep.api)),
                        build.string('previously configured with features'),
                        build.identifier('actual_features'),
                        build.string('but need'),
                        build.identifier('needed_features'),
                    ])
                ]))
            ])),
        ]

    def _create_meson_subdir(self, build: builder.Builder) -> T.List[mparser.BaseNode]:
        # Allow Cargo subprojects to add extra Rust args in meson/meson.build file.
        # This is used to replace build.rs logic.

        # extra_args = []
        # extra_deps = []
        # fs = import('fs')
        # if fs.is_dir('meson')
        #  subdir('meson')
        # endif
        return [
            build.assign(build.array([]), _extra_args_varname()),
            build.assign(build.array([]), _extra_deps_varname()),
            build.assign(build.function('import', [build.string('fs')]), 'fs'),
            build.if_(build.method('is_dir', build.identifier('fs'), [build.string('meson')]),
                      build.block([build.function('subdir', [build.string('meson')])]))
        ]

    def _create_lib(self, pkg: PackageState, build: builder.Builder, crate_type: raw.CRATE_TYPE) -> T.List[mparser.BaseNode]:
        dependencies: T.List[mparser.BaseNode] = []
        dependency_map: T.Dict[mparser.BaseNode, mparser.BaseNode] = {}
        for name in pkg.required_deps:
            dep = pkg.manifest.dependencies[name]
            dependencies.append(build.identifier(_dependency_varname(dep.package)))
            if name != dep.package:
                dep_pkg = self._dep_package(dep)
                dep_lib_name = dep_pkg.manifest.lib.name
                dependency_map[build.string(fixup_meson_varname(dep_lib_name))] = build.string(name)
        for name, sys_dep in pkg.manifest.system_dependencies.items():
            if sys_dep.enabled(pkg.features):
                dependencies.append(build.identifier(f'{fixup_meson_varname(name)}_system_dep'))

        rust_args: T.List[mparser.BaseNode] = [
            build.identifier('features_args'),
            build.identifier(_extra_args_varname()),
            build.identifier('system_deps_args'),
        ]

        dependencies.append(build.identifier(_extra_deps_varname()))

        posargs: T.List[mparser.BaseNode] = [
            build.string(fixup_meson_varname(pkg.manifest.lib.name)),
            build.string(pkg.manifest.lib.path),
        ]

        kwargs: T.Dict[str, mparser.BaseNode] = {
            'dependencies': build.array(dependencies),
            'rust_dependency_map': build.dict(dependency_map),
            'rust_args': build.array(rust_args),
        }

        depname_suffix = '-rs' if crate_type in {'lib', 'rlib', 'proc-macro'} else f'-{crate_type}'
        depname = _dependency_name(pkg.manifest.package.name, pkg.manifest.package.api, depname_suffix)

        lib: mparser.BaseNode
        if pkg.manifest.lib.proc_macro or crate_type == 'proc-macro':
            lib = build.method('proc_macro', build.identifier('rust'), posargs, kwargs)
        else:
            if crate_type in {'lib', 'rlib', 'staticlib'}:
                target_type = 'static_library'
            elif crate_type in {'dylib', 'cdylib'}:
                target_type = 'shared_library'
            else:
                raise MesonException(f'Unsupported crate type {crate_type}')
            if crate_type in {'staticlib', 'cdylib'}:
                kwargs['rust_abi'] = build.string('c')
            lib = build.function(target_type, posargs, kwargs)

        features_args: T.List[mparser.BaseNode] = []
        for f in pkg.features:
            features_args += [build.string('--cfg'), build.string(f'feature="{f}"')]

        # features_args = ['--cfg', 'feature="f1"', ...]
        # lib = xxx_library()
        # dep = declare_dependency()
        # meson.override_dependency()
        return [
            build.assign(build.array(features_args), 'features_args'),
            build.assign(lib, 'lib'),
            build.assign(
                build.function(
                    'declare_dependency',
                    kw={
                        'link_with': build.identifier('lib'),
                        'variables': build.dict({
                            build.string('features'): build.string(','.join(pkg.features)),
                        }),
                        'version': build.string(pkg.manifest.package.version),
                    },
                ),
                'dep'
            ),
            build.method(
                'override_dependency',
                build.identifier('meson'),
                [
                    build.string(depname),
                    build.identifier('dep'),
                ],
            ),
        ]


def load_wraps(source_dir: str, subproject_dir: str) -> T.List[PackageDefinition]:
    """ Convert Cargo.lock into a list of wraps """

    wraps: T.List[PackageDefinition] = []
    filename = os.path.join(source_dir, 'Cargo.lock')
    if os.path.exists(filename):
        try:
            toml = load_toml(filename)
        except TomlImplementationMissing as e:
            mlog.warning('Failed to load Cargo.lock:', str(e), fatal=False)
            return wraps
        raw_cargolock = T.cast('raw.CargoLock', toml)
        cargolock = CargoLock.from_raw(raw_cargolock)
        for package in cargolock.package:
            subp_name = _dependency_name(package.name, version.api(package.version))
            if package.source is None:
                # This is project's package, or one of its workspace members.
                pass
            elif package.source == 'registry+https://github.com/rust-lang/crates.io-index':
                checksum = package.checksum
                if checksum is None:
                    checksum = cargolock.metadata[f'checksum {package.name} {package.version} ({package.source})']
                url = f'https://crates.io/api/v1/crates/{package.name}/{package.version}/download'
                directory = f'{package.name}-{package.version}'
                wraps.append(PackageDefinition.from_values(subp_name, subproject_dir, 'file', {
                    'directory': directory,
                    'source_url': url,
                    'source_filename': f'{directory}.tar.gz',
                    'source_hash': checksum,
                    'method': 'cargo',
                }))
            elif package.source.startswith('git+'):
                parts = urllib.parse.urlparse(package.source[4:])
                query = urllib.parse.parse_qs(parts.query)
                branch = query['branch'][0] if 'branch' in query else ''
                revision = parts.fragment or branch
                url = urllib.parse.urlunparse(parts._replace(params='', query='', fragment=''))
                wraps.append(PackageDefinition.from_values(subp_name, subproject_dir, 'git', {
                    'directory': package.name,
                    'url': url,
                    'revision': revision,
                    'method': 'cargo',
                }))
            else:
                mlog.warning(f'Unsupported source URL in {filename}: {package.source}')
    return wraps



# Copyright 2015-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

"""
Function to check whether the current used LC_CTYPE handles case
transformations of ASCII characters in a way compatible with the POSIX
locale.
"""

import locale
import logging
import multiprocessing
import sys
import textwrap
import traceback

import portage
from portage.util import _unicode_decode, writemsg_level
from portage.util._ctypes import load_libc
from portage.util.futures import asyncio


locale_categories = (
    "LC_COLLATE",
    "LC_CTYPE",
    "LC_MONETARY",
    "LC_MESSAGES",
    "LC_NUMERIC",
    "LC_TIME",
    # GNU extensions
    "LC_ADDRESS",
    "LC_IDENTIFICATION",
    "LC_MEASUREMENT",
    "LC_NAME",
    "LC_PAPER",
    "LC_TELEPHONE",
)

_check_locale_cache = {}


def _check_locale(silent):
    """
    The inner locale check function.
    """
    (libc, _) = load_libc()
    if libc is None:
        return None

    lc = list(range(ord("a"), ord("z") + 1))
    uc = list(range(ord("A"), ord("Z") + 1))
    rlc = [libc.tolower(c) for c in uc]
    ruc = [libc.toupper(c) for c in lc]

    if lc != rlc or uc != ruc:
        if silent:
            return False

        msg = (
            "WARNING: The LC_CTYPE variable is set to a locale "
            + "that specifies transformation between lowercase "
            + "and uppercase ASCII characters that is different than "
            + "the one specified by POSIX locale. This can break "
            + "ebuilds and cause issues in programs that rely on "
            + "the common character conversion scheme. "
            + "Please consider enabling another locale (such as "
            + "en_US.UTF-8) in /etc/locale.gen and setting it "
            + "as LC_CTYPE in make.conf."
        )
        msg = [l for l in textwrap.wrap(msg, 70)]
        msg.append("")
        chars = lambda l: "".join(_unicode_decode(chr(x)) for x in l)
        if uc != ruc:
            msg.extend(
                [
                    f"  {chars(lc)} -> {chars(ruc)}",
                    "  %28s: %s" % ("expected", chars(uc)),
                ]
            )
        if lc != rlc:
            msg.extend(
                [
                    f"  {chars(uc)} -> {chars(rlc)}",
                    "  %28s: %s" % ("expected", chars(lc)),
                ]
            )
        writemsg_level(
            "".join([f"!!! {l}\n" for l in msg]), level=logging.ERROR, noiselevel=-1
        )
        return False

    return True


def _set_and_check_locale(silent, env, mylocale):
    try:
        if env is not None:
            try:
                locale.setlocale(locale.LC_CTYPE, mylocale)
            except locale.Error:
                sys.exit(2)

        ret = _check_locale(silent)
        if ret is None:
            sys.exit(2)
        else:
            sys.exit(0 if ret else 1)
    except Exception:
        traceback.print_exc()
        sys.exit(2)


def check_locale(silent=False, env=None):
    """
    Check whether the locale is sane. Returns True if it is, prints
    warning and returns False if it is not. Returns None if the check
    can not be executed due to platform limitations.
    """
    return asyncio.run(async_check_locale(silent=silent, env=env))


async def async_check_locale(silent=False, env=None):
    if env is not None:
        for v in ("LC_ALL", "LC_CTYPE", "LANG"):
            if v in env:
                mylocale = env[v]
                break
        else:
            mylocale = "C"

        try:
            return _check_locale_cache[mylocale]
        except KeyError:
            pass

    proc = multiprocessing.Process(
        target=_set_and_check_locale,
        args=(silent, env, None if env is None else portage._native_string(mylocale)),
    )
    proc.start()
    proc = portage.process.MultiprocessingProcess(proc)
    await proc.wait()

    pyret = None
    if proc.returncode >= 0:
        ret = proc.returncode
        if ret != 2:
            pyret = ret == 0

    if env is not None:
        _check_locale_cache[mylocale] = pyret
    return pyret


async_check_locale.__doc__ = check_locale.__doc__
async_check_locale.__doc__ += """
    This function is a coroutine.
"""


def split_LC_ALL(env):
    """
    Replace LC_ALL with split-up LC_* variables if it is defined.
    Works on the passed environment (or settings instance).
    """
    lc_all = env.get("LC_ALL")
    if lc_all:
        for c in locale_categories:
            env[c] = lc_all
        # Set empty so that config.reset() can restore LC_ALL state,
        # since del can permanently delete variables which are not
        # stored in the config's backupenv.
        env["LC_ALL"] = ""



import os

_CANDIDATES = [
    # Alpine, Arch, Fedora 34+, OpenWRT, RHEL 9+, FreeBSD, ...
    # It's a symlink on some platforms (Fedora, FreeBSD)
    "/etc/ssl/cert.pem",
    # Fedora <= 34, RHEL <= 9, CentOS <= 9
    "/etc/pki/tls/cert.pem",
    # Debian, Ubuntu (requires ca-certificates)
    "/etc/ssl/certs/ca-certificates.crt",
    # SUSE
    "/etc/ssl/ca-bundle.pem",
]
_SSL_PEM = None


def read_text(_module, _path, encoding="utf-8"):
    with open(where(), "r", encoding=encoding) as data:
        return data.read()


def where():
    global _SSL_PEM
    if _SSL_PEM is not None:
        return _SSL_PEM
    for candidate in _CANDIDATES:
        if os.path.isfile(candidate):
            _SSL_PEM = candidate
            return _SSL_PEM
    else:
        raise FileNotFoundError(", ".join(_CANDIDATES))


def contents():
    return read_text("certifi", "cacert.pem", encoding="utf-8")



# Copyright 2015-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.exception import InvalidData
from portage.localization import _


class NeededEntry:
    """
    Represents one entry (line) from a NEEDED.ELF.2 file. The entry
    must have 5 or more semicolon-delimited fields in order to be
    considered valid. The sixth field is optional, corresponding
    to the multilib category. The multilib_category attribute is
    None if the corresponding field is either empty or missing.
    """

    __slots__ = (
        "arch",
        "filename",
        "multilib_category",
        "needed",
        "runpaths",
        "soname",
    )

    _MIN_FIELDS = 5
    _MULTILIB_CAT_INDEX = 5

    @classmethod
    def parse(cls, filename, line):
        """
        Parse a NEEDED.ELF.2 entry. Raises InvalidData if necessary.

        @param filename: file name for use in exception messages
        @type filename: str
        @param line: a single line of text from a NEEDED.ELF.2 file,
                without a trailing newline
        @type line: str
        @rtype: NeededEntry
        @return: A new NeededEntry instance containing data from line
        """
        fields = line.split(";")
        if len(fields) < cls._MIN_FIELDS:
            raise InvalidData(
                _("Wrong number of fields " "in %s: %s\n\n") % (filename, line)
            )

        obj = cls()
        # Extra fields may exist (for future extensions).
        if len(fields) > cls._MULTILIB_CAT_INDEX and fields[cls._MULTILIB_CAT_INDEX]:
            obj.multilib_category = fields[cls._MULTILIB_CAT_INDEX]
        else:
            obj.multilib_category = None

        del fields[cls._MIN_FIELDS :]
        obj.arch, obj.filename, obj.soname, rpaths, needed = fields
        # We don't use scanelf -q, since that would omit libraries like
        # musl's /usr/lib/libc.so which do not have any DT_NEEDED or
        # DT_SONAME settings. Since we don't use scanelf -q, we have to
        # handle the special rpath value "  -  " below.
        rpaths = "" if rpaths == "  -  " else rpaths
        obj.runpaths = tuple(filter(None, rpaths.split(":")))
        obj.needed = tuple(filter(None, needed.split(",")))

        return obj

    def __str__(self):
        """
        Format this entry for writing to a NEEDED.ELF.2 file.
        """
        return (
            ";".join(
                [
                    self.arch,
                    self.filename,
                    self.soname,
                    ":".join(self.runpaths),
                    ",".join(self.needed),
                    (
                        self.multilib_category
                        if self.multilib_category is not None
                        else ""
                    ),
                ]
            )
            + "\n"
        )



# Copyright 2011 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class RebuildTestCase(TestCase):
    def testRebuild(self):
        """
        Rebuild packages when build-time dependencies are upgraded.
        """

        ebuilds = {
            "sys-libs/x-1": {},
            "sys-libs/x-1-r1": {},
            "sys-libs/x-2": {},
            "sys-apps/a-1": {"DEPEND": "sys-libs/x", "RDEPEND": ""},
            "sys-apps/a-2": {"DEPEND": "sys-libs/x", "RDEPEND": ""},
            "sys-apps/b-1": {"DEPEND": "sys-libs/x", "RDEPEND": ""},
            "sys-apps/b-2": {"DEPEND": "sys-libs/x", "RDEPEND": ""},
            "sys-apps/c-1": {"DEPEND": "sys-libs/x", "RDEPEND": ""},
            "sys-apps/c-2": {"DEPEND": "sys-libs/x", "RDEPEND": ""},
            "sys-apps/d-1": {"RDEPEND": "sys-libs/x"},
            "sys-apps/d-2": {"RDEPEND": "sys-libs/x"},
            "sys-apps/e-2": {"DEPEND": "sys-libs/x", "RDEPEND": ""},
            "sys-apps/f-2": {"DEPEND": "sys-apps/a", "RDEPEND": ""},
            "sys-apps/g-2": {"DEPEND": "sys-apps/b sys-libs/x", "RDEPEND": ""},
        }

        installed = {
            "sys-libs/x-1": {},
            "sys-apps/a-1": {"DEPEND": "sys-libs/x", "RDEPEND": ""},
            "sys-apps/b-1": {"DEPEND": "sys-libs/x", "RDEPEND": ""},
            "sys-apps/c-1": {"DEPEND": "sys-libs/x", "RDEPEND": ""},
            "sys-apps/d-1": {"RDEPEND": "sys-libs/x"},
            "sys-apps/e-1": {"DEPEND": "sys-libs/x", "RDEPEND": ""},
            "sys-apps/f-1": {"DEPEND": "sys-apps/a", "RDEPEND": ""},
            "sys-apps/g-1": {"DEPEND": "sys-apps/b", "RDEPEND": ""},
        }

        world = [
            "sys-apps/a",
            "sys-apps/b",
            "sys-apps/c",
            "sys-apps/d",
            "sys-apps/e",
            "sys-apps/f",
            "sys-apps/g",
        ]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["sys-libs/x", "sys-apps/b"],
                options={
                    "--rebuild-if-unbuilt": True,
                    "--rebuild-exclude": ["sys-apps/c"],
                },
                mergelist=[
                    "sys-libs/x-2",
                    "sys-apps/a-2",
                    "sys-apps/b-2",
                    "sys-apps/e-2",
                    "sys-apps/g-2",
                ],
                ignore_mergelist_order=True,
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["sys-libs/x", "sys-apps/b"],
                options={"--rebuild-if-unbuilt": True},
                mergelist=[
                    "sys-libs/x-2",
                    "sys-apps/a-2",
                    "sys-apps/b-2",
                    "sys-apps/c-2",
                    "sys-apps/e-2",
                    "sys-apps/g-2",
                ],
                ignore_mergelist_order=True,
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["sys-libs/x"],
                options={
                    "--rebuild-if-unbuilt": True,
                    "--rebuild-ignore": ["sys-libs/x"],
                },
                mergelist=["sys-libs/x-2"],
                ignore_mergelist_order=True,
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["sys-libs/x", "sys-apps/b"],
                options={
                    "--rebuild-if-unbuilt": True,
                    "--rebuild-ignore": ["sys-apps/b"],
                },
                mergelist=[
                    "sys-libs/x-2",
                    "sys-apps/a-2",
                    "sys-apps/b-2",
                    "sys-apps/c-2",
                    "sys-apps/e-2",
                ],
                ignore_mergelist_order=True,
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["=sys-libs/x-1-r1", "sys-apps/b"],
                options={"--rebuild-if-unbuilt": True},
                mergelist=[
                    "sys-libs/x-1-r1",
                    "sys-apps/a-2",
                    "sys-apps/b-2",
                    "sys-apps/c-2",
                    "sys-apps/e-2",
                    "sys-apps/g-2",
                ],
                ignore_mergelist_order=True,
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["=sys-libs/x-1-r1", "sys-apps/b"],
                options={"--rebuild-if-new-rev": True},
                mergelist=[
                    "sys-libs/x-1-r1",
                    "sys-apps/a-2",
                    "sys-apps/b-2",
                    "sys-apps/c-2",
                    "sys-apps/e-2",
                    "sys-apps/g-2",
                ],
                ignore_mergelist_order=True,
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["=sys-libs/x-1-r1"],
                options={"--rebuild-if-new-ver": True},
                mergelist=["sys-libs/x-1-r1"],
                ignore_mergelist_order=True,
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["sys-libs/x", "sys-apps/b"],
                options={"--rebuild-if-new-ver": True},
                mergelist=[
                    "sys-libs/x-2",
                    "sys-apps/a-2",
                    "sys-apps/b-2",
                    "sys-apps/c-2",
                    "sys-apps/e-2",
                    "sys-apps/g-2",
                ],
                ignore_mergelist_order=True,
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["=sys-libs/x-1"],
                options={"--rebuild-if-new-rev": True},
                mergelist=["sys-libs/x-1"],
                ignore_mergelist_order=True,
                success=True,
            ),
            ResolverPlaygroundTestCase(
                ["=sys-libs/x-1", "=sys-apps/b-1"],
                options={"--rebuild-if-unbuilt": True},
                mergelist=[
                    "sys-libs/x-1",
                    "sys-apps/a-2",
                    "sys-apps/b-1",
                    "sys-apps/c-2",
                    "sys-apps/e-2",
                    "sys-apps/g-2",
                ],
                ignore_mergelist_order=True,
                success=True,
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# SPDX-License-Identifier: Apache-2.0
# Copyright 2017 The Meson development team


from .base import Dependency, InternalDependency, ExternalDependency, NotFoundDependency, MissingCompiler
from .base import (
        ExternalLibrary, DependencyException, DependencyMethods,
        BuiltinDependency, SystemDependency, get_leaf_external_dependencies)
from .detect import find_external_dependency, get_dep_identifier, packages, _packages_accept_language

__all__ = [
    'Dependency',
    'InternalDependency',
    'ExternalDependency',
    'SystemDependency',
    'BuiltinDependency',
    'NotFoundDependency',
    'ExternalLibrary',
    'DependencyException',
    'DependencyMethods',
    'MissingCompiler',

    'find_external_dependency',
    'get_dep_identifier',
    'get_leaf_external_dependencies',
]

"""Dependency representations and discovery logic.

Meson attempts to largely abstract away dependency discovery information, and
to encapsulate that logic itself so that the DSL doesn't have too much direct
information. There are some cases where this is impossible/undesirable, such
as the `get_variable()` method.

Meson has four primary dependency types:
  1. pkg-config
  2. apple frameworks
  3. CMake
  4. system

Plus a few more niche ones.

When a user calls `dependency('foo')` Meson creates a list of candidates, and
tries those candidates in order to find one that matches the criteria
provided by the user (such as version requirements, or optional components
that are required.)

Except to work around bugs or handle odd corner cases, pkg-config and CMake
generally just work™, though there are exceptions. Most of this package is
concerned with dependencies that don't (always) provide CMake and/or
pkg-config files.

For these cases one needs to write a `system` dependency. These dependencies
descend directly from `ExternalDependency`, in their constructor they
manually set up the necessary link and compile args (and additional
dependencies as necessary).

For example, imagine a dependency called Foo, it uses an environment variable
called `$FOO_ROOT` to point to its install root, which looks like this:
```txt
$FOOROOT
→ include/
→ lib/
```
To use Foo, you need its include directory, and you need to link to
`lib/libfoo.ext`.

You could write code that looks like:

```python
class FooSystemDependency(ExternalDependency):

    def __init__(self, name: str, environment: 'Environment', kwargs: T.Dict[str, T.Any]):
        super().__init__(name, environment, kwargs)
        root = os.environ.get('FOO_ROOT')
        if root is None:
            mlog.debug('$FOO_ROOT is unset.')
            self.is_found = False
            return

        lib = self.clib_compiler.find_library('foo', environment, [os.path.join(root, 'lib')])
        if lib is None:
            mlog.debug('Could not find lib.')
            self.is_found = False
            return

        self.compile_args.append(f'-I{os.path.join(root, "include")}')
        self.link_args.append(lib)
        self.is_found = True
```

This code will look for `FOO_ROOT` in the environment, handle `FOO_ROOT` being
undefined gracefully, then set its `compile_args` and `link_args` gracefully.
It will also gracefully handle not finding the required lib (hopefully that
doesn't happen, but it could if, for example, the lib is only static and
shared linking is requested).

There are a couple of things about this that still aren't ideal. For one, we
don't want to be reading random environment variables at this point. Those
should actually be added to `envconfig.Properties` and read in
`environment.Environment._set_default_properties_from_env` (see how
`BOOST_ROOT` is handled). We can also handle the `static` keyword and the
`prefer_static` built-in option. So now that becomes:

```python
class FooSystemDependency(ExternalDependency):

    def __init__(self, name: str, environment: 'Environment', kwargs: T.Dict[str, T.Any]):
        super().__init__(name, environment, kwargs)
        root = environment.properties[self.for_machine].foo_root
        if root is None:
            mlog.debug('foo_root is unset.')
            self.is_found = False
            return

        get_option = environment.coredata.get_option
        static_opt = kwargs.get('static', get_option(Mesonlib.OptionKey('prefer_static'))
        static = Mesonlib.LibType.STATIC if static_opt else Mesonlib.LibType.SHARED
        lib = self.clib_compiler.find_library(
            'foo', environment, [os.path.join(root, 'lib')], libtype=static)
        if lib is None:
            mlog.debug('Could not find lib.')
            self.is_found = False
            return

        self.compile_args.append(f'-I{os.path.join(root, "include")}')
        self.link_args.append(lib)
        self.is_found = True
```

This is nicer in a couple of ways. First we can properly cross compile as we
are allowed to set `FOO_ROOT` for both the build and host machines, it also
means that users can override this in their machine files, and if that
environment variables changes during a Meson reconfigure Meson won't re-read
it, this is important for reproducibility. Finally, Meson will figure out
whether it should be finding `libfoo.so` or `libfoo.a` (or the platform
specific names). Things are looking pretty good now, so it can be added to
the `packages` dict below:

```python
packages.update({
    'foo': FooSystemDependency,
})
```

Now, what if foo also provides pkg-config, but it's only shipped on Unices,
or only included in very recent versions of the dependency? We can use the
`DependencyFactory` class:

```python
foo_factory = DependencyFactory(
    'foo',
    [DependencyMethods.PKGCONFIG, DependencyMethods.SYSTEM],
    system_class=FooSystemDependency,
)
```

This is a helper function that will generate a default pkg-config based
dependency, and use the `FooSystemDependency` as well. It can also handle
custom finders for pkg-config and cmake based dependencies that need some
extra help. You would then add the `foo_factory` to packages instead of
`FooSystemDependency`:

```python
packages.update({
    'foo': foo_factory,
})
```

If you have a dependency that is very complicated, (such as having multiple
implementations) you may need to write your own factory function. There are a
number of examples in this package.

_Note_ before we moved to factory functions it was common to use an
`ExternalDependency` class that would instantiate different types of
dependencies and hold the one it found. There are a number of drawbacks to
this approach, and no new dependencies should do this.
"""

# This is a dict where the keys should be strings, and the values must be one
# of:
# - An ExternalDependency subclass
# - A DependencyFactory object
# - A callable with a signature of (Environment, MachineChoice, Dict[str, Any]) -> List[Callable[[], ExternalDependency]]
#
# The internal "defaults" attribute contains a separate dictionary mapping
# for lazy imports. The values must be:
# - a string naming the submodule that should be imported from `mesonbuild.dependencies` to populate the dependency
packages.defaults.update({
    # From dev:
    'gtest': 'dev',
    'gmock': 'dev',
    'llvm': 'dev',
    'valgrind': 'dev',
    'zlib': 'dev',
    'jni': 'dev',
    'jdk': 'dev',
    'diasdk': 'dev',

    'boost': 'boost',
    'cuda': 'cuda',

    # per-file
    'coarray': 'coarrays',
    'hdf5': 'hdf5',
    'mpi': 'mpi',
    'scalapack': 'scalapack',

    # From misc:
    'blocks': 'misc',
    'curses': 'misc',
    'netcdf': 'misc',
    'openmp': 'misc',
    'threads': 'misc',
    'pcap': 'misc',
    'cups': 'misc',
    'libwmf': 'misc',
    'libgcrypt': 'misc',
    'gpgme': 'misc',
    'shaderc': 'misc',
    'iconv': 'misc',
    'intl': 'misc',
    'atomic': 'misc',
    'dl': 'misc',
    'openssl': 'misc',
    'libcrypto': 'misc',
    'libssl': 'misc',
    'objfw': 'misc',

    # From platform:
    'appleframeworks': 'platform',

    # from python:
    'numpy': 'python',
    'python3': 'python',
    'pybind11': 'python',

    # From ui:
    'gl': 'ui',
    'gnustep': 'ui',
    'sdl2': 'ui',
    'wxwidgets': 'ui',
    'vulkan': 'ui',

    # from qt
    'qt4': 'qt',
    'qt5': 'qt',
    'qt6': 'qt',
})
_packages_accept_language.update({
    'hdf5',
    'mpi',
    'netcdf',
    'openmp',
})



# Copyright 2010-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ("LocationsManager",)

import shlex
import warnings

import portage
from portage import os, eapi_is_supported, _encodings, _unicode_encode
from portage.const import (
    CUSTOM_PROFILE_PATH,
    GLOBAL_CONFIG_PATH,
    PROFILE_PATH,
    USER_CONFIG_PATH,
)
from portage.eapi import eapi_allows_directories_on_profile_level_and_repository_level
from portage.exception import DirectoryNotFound, InvalidLocation, ParseError
from portage.localization import _
from portage.util import (
    ensure_dirs,
    grabfile,
    normalize_path,
    read_corresponding_eapi_file,
    writemsg,
)
from portage.util._path import exists_raise_eaccess, isdir_raise_eaccess
from portage.repository.config import (
    parse_layout_conf,
    _portage1_profiles_allow_directories,
    _profile_node,
)


_PORTAGE1_DIRECTORIES = frozenset(
    [
        "package.mask",
        "package.provided",
        "package.use",
        "package.use.mask",
        "package.use.force",
        "use.mask",
        "use.force",
    ]
)

_allow_parent_colon = frozenset(["portage-2"])


class LocationsManager:
    def __init__(
        self,
        config_root=None,
        eprefix=None,
        config_profile_path=None,
        local_config=True,
        target_root=None,
        sysroot=None,
    ):
        self.user_profile_dir = None
        self._local_repo_conf_path = None
        self.eprefix = eprefix
        self.config_root = config_root
        self.target_root = target_root
        self.sysroot = sysroot
        self._user_config = local_config

        if self.eprefix is None:
            self.eprefix = portage.const.EPREFIX
        elif self.eprefix:
            self.eprefix = normalize_path(self.eprefix)
            if self.eprefix == os.sep:
                self.eprefix = ""

        if self.config_root is None:
            self.config_root = portage.const.EPREFIX + os.sep

        self.config_root = (
            normalize_path(os.path.abspath(self.config_root or os.sep)).rstrip(os.sep)
            + os.sep
        )

        self._check_var_directory("PORTAGE_CONFIGROOT", self.config_root)
        self.abs_user_config = os.path.join(self.config_root, USER_CONFIG_PATH)
        self.config_profile_path = config_profile_path

        if self.sysroot is None:
            self.sysroot = "/"
        else:
            self.sysroot = (
                normalize_path(os.path.abspath(self.sysroot or os.sep)).rstrip(os.sep)
                + os.sep
            )

        # TODO: Set this via the constructor using
        # PORTAGE_OVERRIDE_EPREFIX.
        self.broot = portage.const.EPREFIX

    def load_profiles(self, repositories, known_repository_paths):
        known_repository_paths = {os.path.realpath(x) for x in known_repository_paths}

        known_repos = []
        for x in known_repository_paths:
            try:
                repo = repositories.get_repo_for_location(x)
            except KeyError:
                layout_data = parse_layout_conf(x)[0]
            else:
                layout_data = {
                    "profile-formats": repo.profile_formats,
                    "profile_eapi_when_unspecified": repo.eapi,
                }
            # force a trailing '/' for ease of doing startswith checks
            known_repos.append((x + "/", layout_data))
        known_repos = tuple(known_repos)

        if self.config_profile_path is None:
            deprecated_profile_path = os.path.join(
                self.config_root, "etc", "make.profile"
            )
            self.config_profile_path = os.path.join(self.config_root, PROFILE_PATH)
            if isdir_raise_eaccess(self.config_profile_path):
                self.profile_path = self.config_profile_path
                if isdir_raise_eaccess(
                    deprecated_profile_path
                ) and not os.path.samefile(self.profile_path, deprecated_profile_path):
                    # Don't warn if they refer to the same path, since
                    # that can be used for backward compatibility with
                    # old software.
                    writemsg(
                        "!!! %s\n"
                        % _("Found 2 make.profile dirs: " "using '%s', ignoring '%s'")
                        % (self.profile_path, deprecated_profile_path),
                        noiselevel=-1,
                    )
            else:
                self.config_profile_path = deprecated_profile_path
                if isdir_raise_eaccess(self.config_profile_path):
                    self.profile_path = self.config_profile_path
                else:
                    self.profile_path = None
        else:
            # NOTE: repoman may pass in an empty string
            # here, in order to create an empty profile
            # for checking dependencies of packages with
            # empty KEYWORDS.
            self.profile_path = self.config_profile_path

        # The symlink might not exist or might not be a symlink.
        self.profiles = []
        self.profiles_complex = []
        if self.profile_path:
            try:
                self._addProfile(
                    os.path.realpath(self.profile_path), repositories, known_repos, ()
                )
            except ParseError as e:
                if not portage._sync_mode:
                    writemsg(
                        _("!!! Unable to parse profile: '%s'\n") % self.profile_path,
                        noiselevel=-1,
                    )
                    writemsg(f"!!! ParseError: {str(e)}\n", noiselevel=-1)
                self.profiles = []
                self.profiles_complex = []

        if self._user_config and self.profiles:
            custom_prof = os.path.join(self.config_root, CUSTOM_PROFILE_PATH)
            if os.path.exists(custom_prof):
                # For read_corresponding_eapi_file, specify default=None
                # in order to allow things like wildcard atoms when
                # is no explicit EAPI setting.
                self.user_profile_dir = custom_prof
                self.profiles.append(custom_prof)
                self.profiles_complex.append(
                    _profile_node(
                        custom_prof,
                        True,
                        True,
                        ("profile-bashrcs", "profile-set"),
                        read_corresponding_eapi_file(
                            custom_prof + os.sep, default=None
                        ),
                        True,
                        show_deprecated_warning=False,
                    )
                )
            del custom_prof

        self.profiles = tuple(self.profiles)
        self.profiles_complex = tuple(self.profiles_complex)

    def _check_var_directory(self, varname, var):
        if not isdir_raise_eaccess(var):
            writemsg(
                _("!!! Error: %s='%s' is not a directory. " "Please correct this.\n")
                % (varname, var),
                noiselevel=-1,
            )
            raise DirectoryNotFound(var)

    def _addProfile(self, currentPath, repositories, known_repos, previous_repos):
        current_abs_path = os.path.abspath(currentPath)
        allow_directories = True
        allow_parent_colon = True
        repo_loc = None
        compat_mode = False
        current_formats = ()
        eapi = None

        intersecting_repos = tuple(
            x for x in known_repos if current_abs_path.startswith(x[0])
        )
        if intersecting_repos:
            # Handle nested repositories. The longest path
            # will be the correct one.
            repo_loc, layout_data = max(intersecting_repos, key=lambda x: len(x[0]))
            eapi = layout_data.get("profile_eapi_when_unspecified")

        eapi_file = os.path.join(currentPath, "eapi")
        eapi = eapi or "0"
        f = None
        try:
            f = open(
                _unicode_encode(eapi_file, encoding=_encodings["fs"], errors="strict"),
                encoding=_encodings["content"],
                errors="replace",
            )
            eapi = f.readline().strip()
        except OSError:
            pass
        else:
            if not eapi_is_supported(eapi):
                raise ParseError(
                    _("Profile contains unsupported " "EAPI '%s': '%s'")
                    % (
                        eapi,
                        os.path.realpath(eapi_file),
                    )
                )
        finally:
            if f is not None:
                f.close()

        if intersecting_repos:
            allow_directories = (
                eapi_allows_directories_on_profile_level_and_repository_level(eapi)
                or any(
                    x in _portage1_profiles_allow_directories
                    for x in layout_data["profile-formats"]
                )
            )
            compat_mode = (
                not eapi_allows_directories_on_profile_level_and_repository_level(eapi)
                and layout_data["profile-formats"] == ("portage-1-compat",)
            )
            allow_parent_colon = any(
                x in _allow_parent_colon for x in layout_data["profile-formats"]
            )
            current_formats = tuple(layout_data["profile-formats"])

        # According to PMS, a deprecated profile warning is not inherited. Since
        # the current profile node may have been inherited by a user profile
        # node, the deprecation warning may be relevant even if it is not a
        # top-level profile node. Therefore, consider the deprecated warning
        # to be irrelevant when the current profile node belongs to the same
        # repo as the previous profile node.
        show_deprecated_warning = tuple(x[0] for x in previous_repos) != tuple(
            x[0] for x in intersecting_repos
        )

        if compat_mode:
            offenders = _PORTAGE1_DIRECTORIES.intersection(os.listdir(currentPath))
            offenders = sorted(
                x for x in offenders if os.path.isdir(os.path.join(currentPath, x))
            )
            if offenders:
                warnings.warn(
                    _(
                        "\nThe selected profile is implicitly using the 'portage-1' format:\n"
                        "\tprofile = %(profile_path)s\n"
                        "But this repository is not using that format:\n"
                        "\trepo = %(repo_name)s\n"
                        "This will break in the future.  Please convert these dirs to files:\n"
                        "\t%(files)s\n"
                        "Or, add this line to the repository's layout.conf:\n"
                        "\tprofile-formats = portage-1"
                    )
                    % dict(
                        profile_path=currentPath,
                        repo_name=repo_loc,
                        files="\n\t".join(offenders),
                    )
                )

        parentsFile = os.path.join(currentPath, "parent")
        if exists_raise_eaccess(parentsFile):
            parents = grabfile(parentsFile)
            if not parents:
                raise ParseError(_("Empty parent file: '%s'") % parentsFile)
            for parentPath in parents:
                abs_parent = parentPath[:1] == os.sep
                if not abs_parent and allow_parent_colon:
                    parentPath = self._expand_parent_colon(
                        parentsFile, parentPath, repo_loc, repositories
                    )

                # NOTE: This os.path.join() call is intended to ignore
                # currentPath if parentPath is already absolute.
                parentPath = normalize_path(os.path.join(currentPath, parentPath))

                if (
                    abs_parent
                    or repo_loc is None
                    or not parentPath.startswith(repo_loc)
                ):
                    # It seems that this parent may point outside
                    # of the current repo, so realpath it.
                    parentPath = os.path.realpath(parentPath)

                if exists_raise_eaccess(parentPath):
                    self._addProfile(
                        parentPath, repositories, known_repos, intersecting_repos
                    )
                else:
                    raise ParseError(
                        _("Parent '%s' not found: '%s'") % (parentPath, parentsFile)
                    )

        self.profiles.append(currentPath)
        self.profiles_complex.append(
            _profile_node(
                currentPath,
                allow_directories,
                False,
                current_formats,
                eapi,
                "build-id" in current_formats,
                show_deprecated_warning=show_deprecated_warning,
            )
        )

    def _expand_parent_colon(self, parentsFile, parentPath, repo_loc, repositories):
        colon = parentPath.find(":")
        if colon == -1:
            return parentPath

        if colon == 0:
            if repo_loc is None:
                raise ParseError(
                    _("Parent '%s' not found: '%s'") % (parentPath, parentsFile)
                )
            else:
                parentPath = normalize_path(
                    os.path.join(repo_loc, "profiles", parentPath[colon + 1 :])
                )
        else:
            p_repo_name = parentPath[:colon]
            try:
                p_repo_loc = repositories.get_location_for_name(p_repo_name)
            except KeyError:
                raise ParseError(
                    _("Parent '%s' not found: '%s'") % (parentPath, parentsFile)
                )
            else:
                parentPath = normalize_path(
                    os.path.join(p_repo_loc, "profiles", parentPath[colon + 1 :])
                )

        return parentPath

    def set_root_override(self, root_overwrite=None):
        # Allow ROOT setting to come from make.conf if it's not overridden
        # by the constructor argument (from the calling environment).
        if self.target_root is None and root_overwrite is not None:
            self.target_root = root_overwrite
            if not self.target_root.strip():
                self.target_root = None
        self.target_root = self.target_root or os.sep

        self.target_root = (
            normalize_path(os.path.abspath(self.target_root)).rstrip(os.path.sep)
            + os.path.sep
        )

        if self.sysroot != "/" and self.target_root == "/":
            writemsg(
                _(
                    "!!! Error: SYSROOT (currently %s) must "
                    "be set to / when ROOT is /.\n"
                )
                % self.sysroot,
                noiselevel=-1,
            )
            raise InvalidLocation(self.sysroot)

        ensure_dirs(self.target_root)
        self._check_var_directory("ROOT", self.target_root)

        self.eroot = self.target_root.rstrip(os.sep) + self.eprefix + os.sep

        # In a cross-prefix scenario where SYSROOT=/ and ROOT=/, assume we want
        # ESYSROOT to point to the target prefix.
        if self.sysroot == self.target_root:
            self.esysroot = self.sysroot.rstrip(os.sep) + self.eprefix + os.sep
        elif self.sysroot == "/":
            self.esysroot = self.broot + os.sep
        else:
            self.esysroot = self.sysroot

        self.global_config_path = GLOBAL_CONFIG_PATH
        if portage.const.EPREFIX:
            self.global_config_path = os.path.join(
                portage.const.EPREFIX, GLOBAL_CONFIG_PATH.lstrip(os.sep)
            )

    def set_port_dirs(self, portdir, portdir_overlay):
        self.portdir = portdir
        self.portdir_overlay = portdir_overlay
        if self.portdir_overlay is None:
            self.portdir_overlay = ""

        self.overlay_profiles = []
        for ov in shlex.split(self.portdir_overlay):
            ov = normalize_path(ov)
            profiles_dir = os.path.join(ov, "profiles")
            if isdir_raise_eaccess(profiles_dir):
                self.overlay_profiles.append(profiles_dir)

        self.profile_locations = [
            os.path.join(portdir, "profiles")
        ] + self.overlay_profiles
        self.profile_and_user_locations = self.profile_locations[:]
        if self._user_config:
            self.profile_and_user_locations.append(self.abs_user_config)

        self.profile_locations = tuple(self.profile_locations)
        self.profile_and_user_locations = tuple(self.profile_and_user_locations)



# Copyright 2014-2024 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class UseFlagsTestCase(TestCase):
    def testUseFlags(self):
        ebuilds = {
            "dev-libs/A-1": {
                "IUSE": "X",
            },
            "dev-libs/B-1": {
                "IUSE": "X Y",
            },
            "dev-libs/C-1": {"IUSE": "abi_x86_32", "EAPI": "7"},
            "dev-libs/D-1": {
                "IUSE": "abi_x86_32",
                "EAPI": "7",
                "RDEPEND": "dev-libs/C[abi_x86_32?]",
            },
        }

        installed = {
            "dev-libs/A-1": {
                "IUSE": "X",
            },
            "dev-libs/B-1": {
                "IUSE": "X",
            },
            "dev-libs/C-1": {"IUSE": "abi_x86_32", "USE": "abi_x86_32", "EAPI": "7"},
            "dev-libs/D-1": {
                "IUSE": "abi_x86_32",
                "USE": "abi_x86_32",
                "EAPI": "7",
                "RDEPEND": "dev-libs/C[abi_x86_32]",
            },
        }

        binpkgs = installed

        user_config = {
            "package.use": (
                "dev-libs/A X",
                "dev-libs/D abi_x86_32",
            ),
            "use.force": ("Y",),
        }

        test_cases = (
            # default: don't reinstall on use flag change
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={"--selective": True, "--usepkg": True},
                success=True,
                mergelist=[],
            ),
            # default: respect use flags for binpkgs
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={"--usepkg": True},
                success=True,
                mergelist=["dev-libs/A-1"],
            ),
            # For bug 773469, we wanted --binpkg-respect-use=y to trigger a
            # slot collision. Instead, a combination of default --autounmask-use
            # combined with --autounmask-backtrack=y from EMERGE_DEFAULT_OPTS
            # triggered this behavior which appeared confusingly similar to
            # --binpkg-respect-use=n behavior.
            # ResolverPlaygroundTestCase(
            # 	["dev-libs/C", "dev-libs/D"],
            # 	options={"--usepkg": True, "--binpkg-respect-use": "y", "--autounmask-backtrack": "y"},
            # 	success=True,
            # 	use_changes={"dev-libs/C-1": {"abi_x86_32": True}},
            # 	mergelist=["[binary]dev-libs/C-1", "[binary]dev-libs/D-1"],
            ResolverPlaygroundTestCase(
                ["dev-libs/C", "dev-libs/D"],
                options={
                    "--usepkg": True,
                    "--binpkg-respect-use": "y",
                    "--autounmask-backtrack": "y",
                },
                success=False,
                slot_collision_solutions=[{"dev-libs/C-1": {"abi_x86_32": True}}],
                mergelist=["dev-libs/C-1", "[binary]dev-libs/D-1"],
            ),
            # --binpkg-respect-use=n: use binpkgs with different use flags
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={"--binpkg-respect-use": "n", "--usepkg": True},
                success=True,
                mergelist=["[binary]dev-libs/A-1"],
            ),
            # --reinstall=changed-use: reinstall if use flag changed
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={"--reinstall": "changed-use", "--usepkg": True},
                success=True,
                mergelist=["dev-libs/A-1"],
            ),
            # --reinstall=changed-use: don't reinstall on new use flag
            ResolverPlaygroundTestCase(
                ["dev-libs/B"],
                options={"--reinstall": "changed-use", "--usepkg": True},
                success=True,
                mergelist=[],
            ),
            # --newuse: reinstall on new use flag
            ResolverPlaygroundTestCase(
                ["dev-libs/B"],
                options={"--newuse": True, "--usepkg": True},
                success=True,
                mergelist=["dev-libs/B-1"],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                user_config["make.conf"] = (f'BINPKG_FORMAT="{binpkg_format}"',)
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    user_config=user_config,
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()

    def testBlockerBinpkgRespectUse(self):
        """
        Test for bug #916336 where we tried to check properties of a blocker
        object which isn't a Package to be merged.
        """

        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "7",
                "IUSE": "abi_x86_32",
                "RDEPEND": "dev-libs/B",
            },
            "dev-libs/B-1": {
                "EAPI": "7",
                "IUSE": "abi_x86_32",
            },
            "dev-libs/A-2": {
                "EAPI": "7",
                "IUSE": "abi_x86_32",
                "RDEPEND": "!<dev-libs/B-2",
            },
            "dev-libs/B-2": {
                "EAPI": "7",
                "IUSE": "abi_x86_32",
            },
        }
        installed = {
            "dev-libs/A-1": {
                "IUSE": "abi_x86_32",
                "USE": "abi_x86_32",
            },
            "dev-libs/B-1": {
                "IUSE": "abi_x86_32",
                "USE": "abi_x86_32",
            },
        }
        binpkgs = ebuilds.copy()

        user_config = {
            "make.conf": (
                'FEATURES="binpkg-multi-instance"',
                'USE="abi_x86_32 abi_x86_32"',
            ),
        }

        world = ("dev-libs/A",)

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A"],
                options={
                    "--verbose": "y",
                    "--update": True,
                    "--deep": True,
                    "--complete-graph": True,
                    "--usepkg": True,
                    "--autounmask": "n",
                    "--autounmask-backtrack": "n",
                    "--autounmask-use": "n",
                },
                success=True,
                mergelist=["dev-libs/A-2", "[uninstall]dev-libs/B-1", "!<dev-libs/B-2"],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                user_config["make.conf"] += (f'BINPKG_FORMAT="{binpkg_format}"',)
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    user_config=user_config,
                    world=world,
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()

    def testNoMergeBinpkgRespectUse(self):
        """
        Testcase for bug #916614 where an incomplete depgraph may be fed into
        _show_ignored_binaries_respect_use.

        We use a mix of +/-abi_x86_32 to trigger the binpkg-respect-use notice
        and depend on a non-existent package in one of the available ebuilds we
        queue to reinstall to trigger an aborted calculation.
        """
        ebuilds = {
            "dev-libs/A-2": {
                "EAPI": "7",
                "IUSE": "abi_x86_32",
            },
            "dev-libs/B-1": {
                "IUSE": "abi_x86_32",
                "RDEPEND": "=dev-libs/A-1",
            },
        }

        installed = {
            "dev-libs/B-1": {
                "IUSE": "abi_x86_32",
                "USE": "abi_x86_32",
            },
            "dev-libs/A-1": {
                "IUSE": "abi_x86_32",
                "USE": "abi_x86_32",
            },
        }

        binpkgs = {
            "dev-libs/A-2": {
                "IUSE": "abi_x86_32",
                "USE": "abi_x86_32",
            },
            "dev-libs/B-1": {
                "IUSE": "abi_x86_32",
                "USE": "",
                "BUILD_ID": "2",
                "BUILD_TIME": "2",
            },
        }

        user_config = {
            "make.conf": (
                'FEATURES="binpkg-multi-instance"',
                'USE="abi_x86_32 abi_x86_32"',
            ),
        }

        world = ("dev-libs/A",)

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@installed"],
                options={
                    "--verbose": "y",
                    "--emptytree": True,
                    "--usepkg": True,
                },
                success=False,
                mergelist=None,
                slot_collision_solutions=None,
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                user_config["make.conf"] += (f'BINPKG_FORMAT="{binpkg_format}"',)
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    user_config=user_config,
                    world=world,
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()



# Copyright 1999-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import functools

from _emerge.CompositeTask import CompositeTask
from portage import os
from portage.dbapi._MergeProcess import MergeProcess
from portage.util._async.AsyncTaskFuture import AsyncTaskFuture


class EbuildMerge(CompositeTask):
    __slots__ = (
        "exit_hook",
        "find_blockers",
        "logger",
        "ldpath_mtimes",
        "pkg",
        "pkg_count",
        "pkg_path",
        "postinst_failure",
        "pretend",
        "settings",
        "tree",
        "world_atom",
    )

    def _start(self):
        root_config = self.pkg.root_config
        settings = self.settings
        mycat = settings["CATEGORY"]
        mypkg = settings["PF"]
        pkgloc = settings["D"]
        infloc = os.path.join(settings["PORTAGE_BUILDDIR"], "build-info")
        myebuild = settings["EBUILD"]
        mydbapi = root_config.trees[self.tree].dbapi
        vartree = root_config.trees["vartree"]
        background = settings.get("PORTAGE_BACKGROUND") == "1"
        logfile = settings.get("PORTAGE_LOG_FILE")

        merge_task = MergeProcess(
            mycat=mycat,
            mypkg=mypkg,
            settings=settings,
            treetype=self.tree,
            vartree=vartree,
            scheduler=self.scheduler,
            background=background,
            blockers=self.find_blockers,
            pkgloc=pkgloc,
            infloc=infloc,
            myebuild=myebuild,
            mydbapi=mydbapi,
            prev_mtimes=self.ldpath_mtimes,
            logfile=logfile,
        )

        self._start_task(merge_task, self._merge_exit)

    def _merge_exit(self, merge_task):
        if self._final_exit(merge_task) != os.EX_OK:
            self._start_exit_hook(self.returncode)
            return

        self.postinst_failure = merge_task.postinst_failure
        pkg = self.pkg
        self.world_atom(pkg)
        pkg_count = self.pkg_count
        pkg_path = self.pkg_path
        logger = self.logger
        if "noclean" not in self.settings.features:
            short_msg = "emerge: ({} of {}) {} Clean Post".format(
                pkg_count.curval,
                pkg_count.maxval,
                pkg.cpv,
            )
            logger.log(
                f" === ({pkg_count.curval} of {pkg_count.maxval}) "
                f"Post-Build Cleaning ({pkg.cpv}::{pkg_path})",
                short_msg=short_msg,
            )
        logger.log(
            f" ::: completed emerge ({pkg_count.curval} of {pkg_count.maxval}) "
            f"{pkg.cpv} to {pkg.root}"
        )

        self._start_exit_hook(self.returncode)

    def _start_exit_hook(self, returncode):
        """
        Start the exit hook, and set returncode after it completes.
        """
        # The returncode will be set after exit hook is complete.
        self.returncode = None
        self._start_task(
            AsyncTaskFuture(future=self.exit_hook(self)),
            functools.partial(self._exit_hook_exit, returncode),
        )

    def _exit_hook_exit(self, returncode, task):
        self._assert_current(task)
        self.returncode = returncode
        self._async_wait()



# Copyright 2007-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import portage.glsa as glsa
from portage._sets.base import PackageSet
from portage.versions import vercmp
from portage._sets import get_boolean

__all__ = ["SecuritySet", "NewGlsaSet", "NewAffectedSet", "AffectedSet"]


class SecuritySet(PackageSet):
    _operations = ["merge"]
    _skip_applied = False

    description = "package set that includes all packages possibly affected by a GLSA"

    def __init__(self, settings, vardbapi, portdbapi, least_change=True):
        super().__init__()
        self._settings = settings
        self._vardbapi = vardbapi
        self._portdbapi = portdbapi
        self._least_change = least_change

    def getGlsaList(self, skip_applied):
        glsaindexlist = glsa.get_glsa_list(self._settings)
        if skip_applied:
            applied_list = glsa.get_applied_glsas(self._settings)
            glsaindexlist = set(glsaindexlist).difference(applied_list)
            glsaindexlist = list(glsaindexlist)
        glsaindexlist.sort()
        return glsaindexlist

    def load(self):
        glsaindexlist = self.getGlsaList(self._skip_applied)
        atomlist = []
        for glsaid in glsaindexlist:
            myglsa = glsa.Glsa(glsaid, self._settings, self._vardbapi, self._portdbapi)
            # print glsaid, myglsa.isVulnerable(), myglsa.isApplied(), myglsa.getMergeList()
            if self.useGlsa(myglsa):
                atomlist += [
                    "=" + x
                    for x in myglsa.getMergeList(least_change=self._least_change)
                ]
        self._setAtoms(self._reduce(atomlist))

    def _reduce(self, atomlist):
        mydict = {}
        for atom in atomlist[:]:
            cpv = self._portdbapi.xmatch("match-all", atom)[0]
            pkg = self._portdbapi._pkg_str(cpv, None)
            cps = f"{pkg.cp}:{pkg.slot}"
            if not cps in mydict:
                mydict[cps] = (atom, cpv)
            else:
                other_cpv = mydict[cps][1]
                if vercmp(cpv.version, other_cpv.version) > 0:
                    atomlist.remove(mydict[cps][0])
                    mydict[cps] = (atom, cpv)
        return atomlist

    def useGlsa(self, myglsa):
        return True

    def updateAppliedList(self):
        glsaindexlist = self.getGlsaList(True)
        applied_list = glsa.get_applied_glsas(self._settings)
        for glsaid in glsaindexlist:
            myglsa = glsa.Glsa(glsaid, self._settings, self._vardbapi, self._portdbapi)
            if not myglsa.isVulnerable() and not myglsa.nr in applied_list:
                myglsa.inject()

    def singleBuilder(cls, options, settings, trees):
        least_change = not get_boolean(options, "use_emerge_resolver", False)
        return cls(
            settings,
            trees["vartree"].dbapi,
            trees["porttree"].dbapi,
            least_change=least_change,
        )

    singleBuilder = classmethod(singleBuilder)


class NewGlsaSet(SecuritySet):
    _skip_applied = True
    description = (
        "Package set that includes all packages possibly affected by an unapplied GLSA"
    )


class AffectedSet(SecuritySet):
    description = "Package set that includes all packages affected by an unapplied GLSA"

    def useGlsa(self, myglsa):
        return myglsa.isVulnerable()


class NewAffectedSet(AffectedSet):
    _skip_applied = True
    description = "Package set that includes all packages affected by an unapplied GLSA"



# SPDX-License-Identifier: Apache-2.0
# Copyright 2019 The Meson development team

from __future__ import annotations
from ntpath import sep as ntsep
from pathlib import Path
from posixpath import sep as posixsep
import hashlib
import os
import typing as T

from . import ExtensionModule, ModuleReturnValue, ModuleInfo
from .. import mlog
from ..build import BuildTarget, CustomTarget, CustomTargetIndex, InvalidArguments
from ..interpreter.type_checking import INSTALL_KW, INSTALL_MODE_KW, INSTALL_TAG_KW, NoneType
from ..interpreterbase import FeatureNew, KwargInfo, typed_kwargs, typed_pos_args, noKwargs
from ..mesonlib import File, MesonException, has_path_sep, is_windows, path_is_in_root, relpath

if T.TYPE_CHECKING:
    from . import ModuleState
    from ..build import BuildTargetTypes
    from ..interpreter import Interpreter
    from ..interpreterbase import TYPE_kwargs
    from ..mesonlib import FileOrString, FileMode

    from typing_extensions import TypedDict

    class ReadKwArgs(TypedDict):
        """Keyword Arguments for fs.read."""

        encoding: str

    class CopyKw(TypedDict):

        """Kwargs for fs.copy"""

        install: bool
        install_dir: T.Optional[str]
        install_mode: FileMode
        install_tag: T.Optional[str]


class FSModule(ExtensionModule):

    INFO = ModuleInfo('fs', '0.53.0')

    def __init__(self, interpreter: Interpreter) -> None:
        super().__init__(interpreter)
        self.methods.update({
            'as_posix': self.as_posix,
            'copyfile': self.copyfile,
            'exists': self.exists,
            'expanduser': self.expanduser,
            'hash': self.hash,
            'is_absolute': self.is_absolute,
            'is_dir': self.is_dir,
            'is_file': self.is_file,
            'is_samepath': self.is_samepath,
            'is_symlink': self.is_symlink,
            'name': self.name,
            'parent': self.parent,
            'read': self.read,
            'relative_to': self.relative_to,
            'replace_suffix': self.replace_suffix,
            'size': self.size,
            'stem': self.stem,
            'suffix': self.suffix,
        })

    def _absolute_dir(self, state: ModuleState, arg: FileOrString) -> str:
        """
        make an absolute path from a relative path, WITHOUT resolving symlinks
        """
        if isinstance(arg, File):
            return arg.absolute_path(state.source_root, state.environment.get_build_dir())
        return os.path.join(state.source_root, state.subdir, os.path.expanduser(arg))

    @staticmethod
    def _obj_to_pathstr(feature_new_prefix: str, obj: T.Union[FileOrString, BuildTargetTypes], state: ModuleState) -> str:
        if isinstance(obj, str):
            return obj

        if isinstance(obj, File):
            FeatureNew(f'{feature_new_prefix} with file', '0.59.0').use(state.subproject, location=state.current_node)
            return str(obj)

        FeatureNew(f'{feature_new_prefix} with build_tgt, custom_tgt, and custom_idx', '1.4.0').use(state.subproject, location=state.current_node)
        return state.backend.get_target_filename(obj)

    def _resolve_dir(self, state: ModuleState, arg: FileOrString) -> str:
        """
        resolves symlinks and makes absolute a directory relative to calling meson.build,
        if not already absolute
        """
        path = self._absolute_dir(state, arg)
        try:
            # accommodate unresolvable paths e.g. symlink loops
            path = os.path.realpath(path)
        except Exception:
            # return the best we could do
            pass
        return path

    @noKwargs
    @FeatureNew('fs.expanduser', '0.54.0')
    @typed_pos_args('fs.expanduser', str)
    def expanduser(self, state: ModuleState, args: T.Tuple[str], kwargs: T.Dict[str, T.Any]) -> str:
        return os.path.expanduser(args[0])

    @noKwargs
    @FeatureNew('fs.is_absolute', '0.54.0')
    @typed_pos_args('fs.is_absolute', (str, File))
    def is_absolute(self, state: ModuleState, args: T.Tuple[FileOrString], kwargs: T.Dict[str, T.Any]) -> bool:
        path = args[0]
        if isinstance(path, File):
            FeatureNew('fs.is_absolute with file', '0.59.0').use(state.subproject, location=state.current_node)
            path = str(path)
        if is_windows():
            # os.path.isabs was broken for Windows before Python 3.13, so we implement it ourselves
            path = path[:3].replace(posixsep, ntsep)
            return path.startswith(ntsep * 2) or path.startswith(':' + ntsep, 1)
        return path.startswith(posixsep)

    @noKwargs
    @FeatureNew('fs.as_posix', '0.54.0')
    @typed_pos_args('fs.as_posix', str)
    def as_posix(self, state: ModuleState, args: T.Tuple[str], kwargs: T.Dict[str, T.Any]) -> str:
        r"""
        this function assumes you are passing a Windows path, even if on a Unix-like system
        and so ALL '\' are turned to '/', even if you meant to escape a character
        """
        return args[0].replace(ntsep, posixsep)

    @noKwargs
    @typed_pos_args('fs.exists', str)
    def exists(self, state: ModuleState, args: T.Tuple[str], kwargs: T.Dict[str, T.Any]) -> bool:
        return os.path.exists(self._resolve_dir(state, args[0]))

    @noKwargs
    @typed_pos_args('fs.is_symlink', (str, File))
    def is_symlink(self, state: ModuleState, args: T.Tuple[FileOrString], kwargs: T.Dict[str, T.Any]) -> bool:
        if isinstance(args[0], File):
            FeatureNew('fs.is_symlink with file', '0.59.0').use(state.subproject, location=state.current_node)
        return os.path.islink(self._absolute_dir(state, args[0]))

    @noKwargs
    @typed_pos_args('fs.is_file', str)
    def is_file(self, state: ModuleState, args: T.Tuple[str], kwargs: T.Dict[str, T.Any]) -> bool:
        return os.path.isfile(self._resolve_dir(state, args[0]))

    @noKwargs
    @typed_pos_args('fs.is_dir', str)
    def is_dir(self, state: ModuleState, args: T.Tuple[str], kwargs: T.Dict[str, T.Any]) -> bool:
        return os.path.isdir(self._resolve_dir(state, args[0]))

    @noKwargs
    @typed_pos_args('fs.hash', (str, File), str)
    def hash(self, state: ModuleState, args: T.Tuple[FileOrString, str], kwargs: T.Dict[str, T.Any]) -> str:
        if isinstance(args[0], File):
            FeatureNew('fs.hash with file', '0.59.0').use(state.subproject, location=state.current_node)
        file = self._resolve_dir(state, args[0])
        if not os.path.isfile(file):
            raise MesonException(f'{file} is not a file and therefore cannot be hashed')
        try:
            h = hashlib.new(args[1])
        except ValueError:
            raise MesonException('hash algorithm {} is not available'.format(args[1]))
        mlog.debug('computing {} sum of {} size {} bytes'.format(args[1], file, os.stat(file).st_size))
        with open(file, mode='rb', buffering=0) as f:
            h.update(f.read())
        return h.hexdigest()

    @noKwargs
    @typed_pos_args('fs.size', (str, File))
    def size(self, state: ModuleState, args: T.Tuple[FileOrString], kwargs: T.Dict[str, T.Any]) -> int:
        if isinstance(args[0], File):
            FeatureNew('fs.size with file', '0.59.0').use(state.subproject, location=state.current_node)
        file = self._resolve_dir(state, args[0])
        if not os.path.isfile(file):
            raise MesonException(f'{file} is not a file and therefore cannot be sized')
        try:
            return os.stat(file).st_size
        except ValueError:
            raise MesonException('{} size could not be determined'.format(args[0]))

    @noKwargs
    @typed_pos_args('fs.is_samepath', (str, File), (str, File))
    def is_samepath(self, state: ModuleState, args: T.Tuple[FileOrString, FileOrString], kwargs: T.Dict[str, T.Any]) -> bool:
        if isinstance(args[0], File) or isinstance(args[1], File):
            FeatureNew('fs.is_samepath with file', '0.59.0').use(state.subproject, location=state.current_node)
        file1 = self._resolve_dir(state, args[0])
        file2 = self._resolve_dir(state, args[1])
        if not os.path.exists(file1):
            return False
        if not os.path.exists(file2):
            return False
        try:
            return os.path.samefile(file1, file2)
        except OSError:
            return False

    @noKwargs
    @typed_pos_args('fs.replace_suffix', (str, File, CustomTarget, CustomTargetIndex, BuildTarget), str)
    def replace_suffix(self, state: ModuleState, args: T.Tuple[T.Union[FileOrString, BuildTargetTypes], str], kwargs: T.Dict[str, T.Any]) -> str:
        if args[1] and not args[1].startswith('.'):
            raise ValueError(f"Invalid suffix {args[1]!r}")
        path = self._obj_to_pathstr('fs.replace_suffix', args[0], state)
        return os.path.splitext(path)[0] + args[1]

    @noKwargs
    @typed_pos_args('fs.parent', (str, File, CustomTarget, CustomTargetIndex, BuildTarget))
    def parent(self, state: ModuleState, args: T.Tuple[T.Union[FileOrString, BuildTargetTypes]], kwargs: T.Dict[str, T.Any]) -> str:
        path = self._obj_to_pathstr('fs.parent', args[0], state)
        return os.path.split(path)[0] or '.'

    @noKwargs
    @typed_pos_args('fs.name', (str, File, CustomTarget, CustomTargetIndex, BuildTarget))
    def name(self, state: ModuleState, args: T.Tuple[T.Union[FileOrString, BuildTargetTypes]], kwargs: T.Dict[str, T.Any]) -> str:
        path = self._obj_to_pathstr('fs.name', args[0], state)
        return os.path.basename(path)

    @noKwargs
    @typed_pos_args('fs.stem', (str, File, CustomTarget, CustomTargetIndex, BuildTarget))
    @FeatureNew('fs.stem', '0.54.0')
    def stem(self, state: ModuleState, args: T.Tuple[T.Union[FileOrString, BuildTargetTypes]], kwargs: T.Dict[str, T.Any]) -> str:
        path = self._obj_to_pathstr('fs.name', args[0], state)
        return os.path.splitext(os.path.basename(path))[0]

    @noKwargs
    @typed_pos_args('fs.suffix', (str, File, CustomTarget, CustomTargetIndex, BuildTarget))
    @FeatureNew('fs.suffix', '1.9.0')
    def suffix(self, state: ModuleState, args: T.Tuple[T.Union[FileOrString, BuildTargetTypes]], kwargs: T.Dict[str, T.Any]) -> str:
        path = self._obj_to_pathstr('fs.suffix', args[0], state)
        return os.path.splitext(path)[1]

    @FeatureNew('fs.read', '0.57.0')
    @typed_pos_args('fs.read', (str, File))
    @typed_kwargs('fs.read', KwargInfo('encoding', str, default='utf-8'))
    def read(self, state: ModuleState, args: T.Tuple[FileOrString], kwargs: ReadKwArgs) -> str:
        """Read a file from the source tree and return its value as a decoded
        string.

        If the encoding is not specified, the file is assumed to be utf-8
        encoded. Paths must be relative by default (to prevent accidents) and
        are forbidden to be read from the build directory (to prevent build
        loops)
        """
        path = args[0]
        encoding = kwargs['encoding']
        src_dir = state.environment.source_dir
        sub_dir = state.subdir
        build_dir = state.environment.get_build_dir()

        if isinstance(path, File):
            if path.is_built:
                raise MesonException(
                    'fs.read does not accept built files() objects')
            path = os.path.join(src_dir, path.relative_name())
        else:
            if sub_dir:
                src_dir = os.path.join(src_dir, sub_dir)
            path = os.path.join(src_dir, path)

        path = os.path.abspath(path)
        if path_is_in_root(Path(path), Path(build_dir), resolve=True):
            raise MesonException('path must not be in the build tree')
        try:
            with open(path, encoding=encoding) as f:
                data = f.read()
        except FileNotFoundError:
            raise MesonException(f'File {args[0]} does not exist.')
        except UnicodeDecodeError:
            raise MesonException(f'decoding failed for {args[0]}')
        # Reconfigure when this file changes as it can contain data used by any
        # part of the build configuration (e.g. `project(..., version:
        # fs.read_file('VERSION')` or `configure_file(...)`
        self.interpreter.add_build_def_file(path)
        return data

    @FeatureNew('fs.copyfile', '0.64.0')
    @typed_pos_args('fs.copyfile', (File, str), optargs=[str])
    @typed_kwargs(
        'fs.copyfile',
        INSTALL_KW,
        INSTALL_MODE_KW,
        INSTALL_TAG_KW,
        KwargInfo('install_dir', (str, NoneType)),
    )
    def copyfile(self, state: ModuleState, args: T.Tuple[FileOrString, T.Optional[str]],
                 kwargs: CopyKw) -> ModuleReturnValue:
        """Copy a file into the build directory at build time."""
        if kwargs['install'] and not kwargs['install_dir']:
            raise InvalidArguments('"install_dir" must be specified when "install" is true')

        src = self.interpreter.source_strings_to_files([args[0]])[0]

        # The input is allowed to have path separators, but the output may not,
        # so use the basename for the default case
        dest = args[1] if args[1] else os.path.basename(src.fname)
        if has_path_sep(dest):
            raise InvalidArguments('Destination path may not have path separators')

        ct = CustomTarget(
            dest,
            state.subdir,
            state.subproject,
            state.environment,
            state.environment.get_build_command() + ['--internal', 'copy', '@INPUT@', '@OUTPUT@'],
            [src],
            [dest],
            build_by_default=True,
            install=kwargs['install'],
            install_dir=[kwargs['install_dir']],
            install_mode=kwargs['install_mode'],
            install_tag=[kwargs['install_tag']],
            backend=state.backend,
            description='Copying file {}',
        )

        return ModuleReturnValue(ct, [ct])

    @FeatureNew('fs.relative_to', '1.3.0')
    @typed_pos_args('fs.relative_to', (str, File, CustomTarget, CustomTargetIndex, BuildTarget), (str, File, CustomTarget, CustomTargetIndex, BuildTarget))
    @noKwargs
    def relative_to(self, state: ModuleState, args: T.Tuple[T.Union[FileOrString, BuildTargetTypes], T.Union[FileOrString, BuildTargetTypes]], kwargs: TYPE_kwargs) -> str:
        def to_path(arg: T.Union[FileOrString, CustomTarget, CustomTargetIndex, BuildTarget]) -> str:
            if isinstance(arg, File):
                return arg.absolute_path(state.environment.source_dir, state.environment.build_dir)
            elif isinstance(arg, (CustomTarget, CustomTargetIndex, BuildTarget)):
                return state.backend.get_target_filename_abs(arg)
            else:
                return os.path.join(state.environment.source_dir, state.subdir, arg)

        t = to_path(args[0])
        f = to_path(args[1])

        return relpath(t, f)


def initialize(*args: T.Any, **kwargs: T.Any) -> FSModule:
    return FSModule(*args, **kwargs)



# Copyright 2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import pytest

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class BinpackageDowngradesSlotDepTestCase(TestCase):
    @pytest.mark.xfail()
    def testBinpackageDowngradesSlotDep(self):
        python_use = "python_targets_python3_12 +python_targets_python3_13"
        python_usedep = "python_targets_python3_12(-)?,python_targets_python3_13(-)?"
        common_metadata = {
            "EAPI": "8",
            "REQUIRED_USE": """
                python? (
                    || (
                        python_targets_python3_12
                        python_targets_python3_13
                    )
                )
            """,
        }

        ebuilds = {
            "dev-libs/libxml2-2.13.9": {
                "IUSE": f"+python {python_use}",
                "SLOT": "2",
                **common_metadata,
            },
            "dev-libs/libxml2-2.14.6": {
                "IUSE": f"+python {python_use}",
                "SLOT": "2/16",
                **common_metadata,
            },
            "dev-libs/libxslt-1.1.43-r1": {
                "IUSE": f"python {python_use}",
                "RDEPEND": f"""
                    >=dev-libs/libxml2-2.13:2=
                    python? (
                        >=dev-libs/libxml2-2.13:2=[python,{python_usedep}]
                    )
                """,
                **common_metadata,
            },
        }

        binpkgs = {
            "dev-libs/libxslt-1.1.43-r1": {
                "IUSE": f"python {python_use}",
                "RDEPEND": f"""
                    >=dev-libs/libxml2-2.13:2/2=
                    python? (
                        >=dev-libs/libxml2-2.13:2/2=[python,{python_usedep}]
                    )
                """,
                **common_metadata,
            },
        }

        installed = {
            "dev-libs/libxml2-2.14.6": {
                "IUSE": f"+python {python_use}",
                "USE": f"python python_targets_python3_13",
                "SLOT": "2/16",
                **common_metadata,
            },
            "dev-libs/libxslt-1.1.43-r1": {
                "IUSE": f"python {python_use}",
                "USE": "python_targets_python3_13",
                "RDEPEND": f"""
                    >=dev-libs/libxml2-2.13:2/16=
                    python? (
                        >=dev-libs/libxml2-2.13:2/16=[python,{python_usedep}]
                    )
                """,
                **common_metadata,
            },
        }

        world = []
        user_config = {"package.use": ["*/* -python_targets_python3_13"]}

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            installed=installed,
            binpkgs=binpkgs,
            world=world,
            user_config=user_config,
            debug=False,
        )

        settings = playground.settings
        profile_path = settings.profile_path

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/libxslt"],
                success=True,
                options={"--usepkg": True},
                mergelist=["dev-libs/libxslt-1.1.43-r1"],
            ),
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



import difflib
from email.headerregistry import Address
import errno
import logging
import os
import os.path as osp
from pathlib import Path
import re

try:
    import tomllib
except ImportError:
    try:
        import tomli as tomllib
    # Some downstream distributors remove the vendored tomli.
    # When that is removed, import tomli from the regular location.
    except ImportError:
        import tomli as tomllib

from ._spdx_data import licenses
from .common import normalise_core_metadata_name
from .versionno import normalise_version

log = logging.getLogger(__name__)


class ConfigError(ValueError):
    pass

metadata_list_fields = {
    'classifiers',
    'requires',
    'dev-requires'
}

metadata_allowed_fields = {
    'module',
    'author',
    'author-email',
    'maintainer',
    'maintainer-email',
    'home-page',
    'license',
    'keywords',
    'requires-python',
    'dist-name',
    'description-file',
    'requires-extra',
} | metadata_list_fields

metadata_required_fields = {
    'module',
    'author',
}

pep621_allowed_fields = {
    'name',
    'version',
    'description',
    'readme',
    'requires-python',
    'license',
    'license-files',
    'authors',
    'maintainers',
    'keywords',
    'classifiers',
    'urls',
    'scripts',
    'gui-scripts',
    'entry-points',
    'dependencies',
    'optional-dependencies',
    'dynamic',
}

default_license_files_globs = ['COPYING*', 'LICEN[CS]E*']
license_files_allowed_chars = re.compile(r'^[\w\-\.\/\*\?\[\]]+$')


def read_flit_config(path):
    """Read and check the `pyproject.toml` file with data about the package.
    """
    d = tomllib.loads(path.read_text('utf-8'))
    return prep_toml_config(d, path)


class EntryPointsConflict(ConfigError):
    def __str__(self):
        return ('Please specify console_scripts entry points, or [scripts] in '
            'flit config, not both.')

def prep_toml_config(d, path):
    """Validate config loaded from pyproject.toml and prepare common metadata

    Returns a LoadedConfig object.
    """
    dtool = d.get('tool', {}).get('flit', {})

    if 'project' in d:
        # Metadata in [project] table (PEP 621)
        if 'metadata' in dtool:
            raise ConfigError(
                "Use [project] table for metadata or [tool.flit.metadata], not both."
            )
        if ('scripts' in dtool) or ('entrypoints' in dtool):
            raise ConfigError(
                "Don't mix [project] metadata with [tool.flit.scripts] or "
                "[tool.flit.entrypoints]. Use [project.scripts],"
                "[project.gui-scripts] or [project.entry-points] as replacements."
            )
        loaded_cfg = read_pep621_metadata(d['project'], path)

        module_tbl = dtool.get('module', {})
        if 'name' in module_tbl:
            loaded_cfg.module = module_tbl['name']
    elif 'metadata' in dtool:
        # Metadata in [tool.flit.metadata] (pre PEP 621 format)
        if 'module' in dtool:
            raise ConfigError(
                "Use [tool.flit.module] table with new-style [project] metadata, "
                "not [tool.flit.metadata]"
            )
        loaded_cfg = _prep_metadata(dtool['metadata'], path)
        loaded_cfg.dynamic_metadata = ['version', 'description']

        if 'entrypoints' in dtool:
            loaded_cfg.entrypoints = flatten_entrypoints(dtool['entrypoints'])

        if 'scripts' in dtool:
            loaded_cfg.add_scripts(dict(dtool['scripts']))
    else:
        raise ConfigError(
            "Neither [project] nor [tool.flit.metadata] found in pyproject.toml"
        )

    unknown_sections = set(dtool) - {
        'metadata', 'module', 'scripts', 'entrypoints', 'sdist', 'external-data'
    }
    unknown_sections = [s for s in unknown_sections if not s.lower().startswith('x-')]
    if unknown_sections:
        raise ConfigError('Unexpected tables in pyproject.toml: ' + ', '.join(
            '[tool.flit.{}]'.format(s) for s in unknown_sections
        ))

    if 'sdist' in dtool:
        unknown_keys = set(dtool['sdist']) - {'include', 'exclude'}
        if unknown_keys:
            raise ConfigError(
                "Unknown keys in [tool.flit.sdist]:" + ", ".join(unknown_keys)
            )

        loaded_cfg.sdist_include_patterns = _check_glob_patterns(
            dtool['sdist'].get('include', []), 'include'
        )
        exclude = [
            "**/__pycache__",
            "**.pyc",
        ] + dtool['sdist'].get('exclude', [])
        loaded_cfg.sdist_exclude_patterns = _check_glob_patterns(
            exclude, 'exclude'
        )

    data_dir = dtool.get('external-data', {}).get('directory', None)
    if data_dir is not None:
        toml_key = "tool.flit.external-data.directory"
        if not isinstance(data_dir, str):
            raise ConfigError(f"{toml_key} must be a string")

        normp = osp.normpath(data_dir)
        if isabs_ish(normp):
            raise ConfigError(f"{toml_key} cannot be an absolute path")
        if normp.startswith('..' + os.sep):
            raise ConfigError(
                f"{toml_key} cannot point outside the directory containing pyproject.toml"
            )
        if normp == '.':
            raise ConfigError(
                f"{toml_key} cannot refer to the directory containing pyproject.toml"
            )
        loaded_cfg.data_directory = path.parent / data_dir
        if not loaded_cfg.data_directory.is_dir():
            raise ConfigError(f"{toml_key} must refer to a directory")

    return loaded_cfg

def flatten_entrypoints(ep):
    """Flatten nested entrypoints dicts.

    Entry points group names can include dots. But dots in TOML make nested
    dictionaries:

    [entrypoints.a.b]    # {'entrypoints': {'a': {'b': {}}}}

    The proper way to avoid this is:

    [entrypoints."a.b"]  # {'entrypoints': {'a.b': {}}}

    But since there isn't a need for arbitrarily nested mappings in entrypoints,
    flit allows you to use the former. This flattens the nested dictionaries
    from loading pyproject.toml.
    """
    def _flatten(d, prefix):
        d1 = {}
        for k, v in d.items():
            if isinstance(v, dict):
                for flattened in _flatten(v, prefix+'.'+k):
                    yield flattened
            else:
                d1[k] = v

        if d1:
            yield prefix, d1

    res = {}
    for k, v in ep.items():
        res.update(_flatten(v, k))
    return res


def _check_glob_patterns(pats, clude):
    """Check and normalise glob patterns for sdist include/exclude"""
    if not isinstance(pats, list):
        raise ConfigError("sdist {} patterns must be a list".format(clude))

    # Windows filenames can't contain these (nor * or ?, but they are part of
    # glob patterns) - https://stackoverflow.com/a/31976060/434217
    bad_chars = re.compile(r'[\000-\037<>:"\\]')

    normed = []

    for p in pats:
        if bad_chars.search(p):
            raise ConfigError(
                '{} pattern {!r} contains bad characters (<>:\"\\ or control characters)'
                .format(clude, p)
            )

        normp = osp.normpath(p)

        if isabs_ish(normp):
            raise ConfigError(
                f'{clude} pattern {p!r} is an absolute path'
            )
        if normp.startswith('..' + os.sep):
            raise ConfigError(
                '{} pattern {!r} points out of the directory containing pyproject.toml'
                .format(clude, p)
            )
        normed.append(normp)

    return normed


class LoadedConfig(object):
    def __init__(self):
        self.module = None
        self.metadata = {}
        self.reqs_by_extra = {}
        self.entrypoints = {}
        self.referenced_files = []
        self.sdist_include_patterns = []
        self.sdist_exclude_patterns = []
        self.dynamic_metadata = []
        self.data_directory = None

    def add_scripts(self, scripts_dict):
        if scripts_dict:
            if 'console_scripts' in self.entrypoints:
                raise EntryPointsConflict
            else:
                self.entrypoints['console_scripts'] = scripts_dict

readme_ext_to_content_type = {
    '.rst': 'text/x-rst',
    '.md': 'text/markdown',
    '.txt': 'text/plain',
}


def description_from_file(rel_path: str, proj_dir: Path, guess_mimetype=True):
    if isabs_ish(rel_path):
        raise ConfigError("Readme path must be relative")

    desc_path = proj_dir / rel_path
    try:
        with desc_path.open('r', encoding='utf-8') as f:
            raw_desc = f.read()
    except IOError as e:
        if e.errno == errno.ENOENT:
            raise ConfigError(
                "Description file {} does not exist".format(desc_path)
            )
        raise

    if guess_mimetype:
        ext = desc_path.suffix.lower()
        try:
            mimetype = readme_ext_to_content_type[ext]
        except KeyError:
            log.warning("Unknown extension %r for description file.", ext)
            log.warning("  Recognised extensions: %s",
                        " ".join(readme_ext_to_content_type))
            mimetype = None
    else:
        mimetype = None

    return raw_desc, mimetype


def _prep_metadata(md_sect, path):
    """Process & verify the metadata from a config file

    - Pull out the module name we're packaging.
    - Read description-file and check that it's valid rst
    - Convert dashes in key names to underscores
      (e.g. home-page in config -> home_page in metadata)
    """
    if not set(md_sect).issuperset(metadata_required_fields):
        missing = metadata_required_fields - set(md_sect)
        raise ConfigError("Required fields missing: " + '\n'.join(missing))

    res = LoadedConfig()

    res.module = md_sect.get('module')
    if not all([m.isidentifier() for m in res.module.split(".")]):
        raise ConfigError("Module name %r is not a valid identifier" % res.module)

    md_dict = res.metadata

    # Description file
    if 'description-file' in md_sect:
        desc_path = md_sect.get('description-file')
        res.referenced_files.append(desc_path)
        desc_content, mimetype = description_from_file(desc_path, path.parent)
        md_dict['description'] =  desc_content
        md_dict['description_content_type'] = mimetype

    if 'urls' in md_sect:
        project_urls = md_dict['project_urls'] = []
        for label, url in sorted(md_sect.pop('urls').items()):
            project_urls.append("{}, {}".format(label, url))

    for key, value in md_sect.items():
        if key in {'description-file', 'module'}:
            continue
        if key not in metadata_allowed_fields:
            closest = difflib.get_close_matches(key, metadata_allowed_fields,
                                                n=1, cutoff=0.7)
            msg = "Unrecognised metadata key: {!r}".format(key)
            if closest:
                msg += " (did you mean {!r}?)".format(closest[0])
            raise ConfigError(msg)

        k2 = key.replace('-', '_')
        md_dict[k2] = value
        if key in metadata_list_fields:
            if not isinstance(value, list):
                raise ConfigError('Expected a list for {} field, found {!r}'
                                    .format(key, value))
            if not all(isinstance(a, str) for a in value):
                raise ConfigError('Expected a list of strings for {} field'
                                    .format(key))
        elif key == 'requires-extra':
            if not isinstance(value, dict):
                raise ConfigError('Expected a dict for requires-extra field, found {!r}'
                                    .format(value))
            if not all(isinstance(e, list) for e in value.values()):
                raise ConfigError('Expected a dict of lists for requires-extra field')
            for e, reqs in value.items():
                if not all(isinstance(a, str) for a in reqs):
                    raise ConfigError('Expected a string list for requires-extra. (extra {})'
                                        .format(e))
        else:
            if not isinstance(value, str):
                raise ConfigError('Expected a string for {} field, found {!r}'
                                    .format(key, value))

    # What we call requires in the ini file is technically requires_dist in
    # the metadata.
    if 'requires' in md_dict:
        md_dict['requires_dist'] = md_dict.pop('requires')

    # And what we call dist-name is name in the metadata
    if 'dist_name' in md_dict:
        md_dict['name'] = md_dict.pop('dist_name')

    # Move dev-requires into requires-extra
    reqs_noextra = md_dict.pop('requires_dist', [])

    reqs_extra = md_dict.pop('requires_extra', {})
    extra_names_by_normed = {}
    for e, reqs in reqs_extra.items():
        if not all(isinstance(a, str) for a in reqs):
            raise ConfigError(
                f'Expected a string list for requires-extra group {e}'
            )
        if not name_is_valid(e):
            raise ConfigError(
                f'requires-extra group name {e!r} is not valid'
            )
        enorm = normalise_core_metadata_name(e)
        extra_names_by_normed.setdefault(enorm, set()).add(e)
        res.reqs_by_extra[enorm] = reqs

    clashing_extra_names = [
        g for g in extra_names_by_normed.values() if len(g) > 1
    ]
    if clashing_extra_names:
        fmted = ['/'.join(sorted(g)) for g in clashing_extra_names]
        raise ConfigError(
            f"requires-extra group names clash: {'; '.join(fmted)}"
        )

    dev_requires = md_dict.pop('dev_requires', None)
    if dev_requires is not None:
        if 'dev' in res.reqs_by_extra:
            raise ConfigError(
                'dev-requires occurs together with its replacement requires-extra.dev.')
        else:
            log.warning(
                '"dev-requires = ..." is obsolete. Use "requires-extra = {"dev" = ...}" instead.')
            res.reqs_by_extra['dev'] = dev_requires

    # Add requires-extra requirements into requires_dist
    md_dict['requires_dist'] = \
        reqs_noextra + list(_expand_requires_extra(res.reqs_by_extra))

    md_dict['provides_extra'] = sorted(res.reqs_by_extra.keys())

    # For internal use, record the main requirements as a '.none' extra.
    res.reqs_by_extra['.none'] = reqs_noextra

    if path:
        license_files = sorted(
            _license_files_from_globs(
                path.parent, default_license_files_globs, warn_no_files=False
            )
        )
        res.referenced_files.extend(license_files)
        md_dict['license_files'] = license_files

    return res

def _expand_requires_extra(re):
    for extra, reqs in sorted(re.items()):
        for req in reqs:
            if ';' in req:
                name, envmark = req.split(';', 1)
                yield '{} ; extra == "{}" and ({})'.format(name, extra, envmark)
            else:
                yield '{} ; extra == "{}"'.format(req, extra)


def _license_files_from_globs(project_dir: Path, globs, warn_no_files = True):
    license_files = set()
    for pattern in globs:
        if isabs_ish(pattern):
            raise ConfigError(
                "Invalid glob pattern for [project.license-files]: '{}'. "
                "Pattern must not start with '/'.".format(pattern)
            )
        if ".." in pattern:
            raise ConfigError(
                "Invalid glob pattern for [project.license-files]: '{}'. "
                "Pattern must not contain '..'".format(pattern)
            )
        if license_files_allowed_chars.match(pattern) is None:
            raise ConfigError(
                "Invalid glob pattern for [project.license-files]: '{}'. "
                "Pattern contains invalid characters. "
                "https://packaging.python.org/en/latest/specifications/pyproject-toml/#license-files"
            )
        try:
            files = [
                file.relative_to(project_dir).as_posix()
                for file in project_dir.glob(pattern)
                if file.is_file()
            ]
        except ValueError as ex:
            raise ConfigError(
                "Invalid glob pattern for [project.license-files]: '{}'. {}".format(pattern, ex.args[0])
            )

        if not files and warn_no_files:
            raise ConfigError(
                "No files found for [project.license-files]: '{}' pattern".format(pattern)
            )
        license_files.update(files)
    return license_files

def _check_type(d, field_name, cls):
    if not isinstance(d[field_name], cls):
        raise ConfigError(
            "{} field should be {}, not {}".format(field_name, cls, type(d[field_name]))
        )

def _check_types(d, field_name, cls_list) -> None:
    if not isinstance(d[field_name], cls_list):
        raise ConfigError(
            "{} field should be {}, not {}".format(
                field_name, ' or '.join(map(str, cls_list)), type(d[field_name])
            )
        )

def _check_list_of_str(d, field_name):
    if not isinstance(d[field_name], list) or not all(
        isinstance(e, str) for e in d[field_name]
    ):
        raise ConfigError(
            "{} field should be a list of strings".format(field_name)
        )

def read_pep621_metadata(proj, path) -> LoadedConfig:
    lc = LoadedConfig()
    md_dict = lc.metadata

    if 'name' not in proj:
        raise ConfigError('name must be specified in [project] table')
    _check_type(proj, 'name', str)
    if not name_is_valid(proj['name']):
        raise ConfigError(f"name {proj['name']} is not valid")
    md_dict['name'] = proj['name']
    lc.module = md_dict['name'].replace('-', '_')

    unexpected_keys = proj.keys() - pep621_allowed_fields
    if unexpected_keys:
        log.warning("Unexpected names under [project]: %s", ', '.join(unexpected_keys))

    if 'version' in proj:
        _check_type(proj, 'version', str)
        md_dict['version'] = normalise_version(proj['version'])
    if 'description' in proj:
        _check_type(proj, 'description', str)
        md_dict['summary'] = proj['description']
    if 'readme' in proj:
        readme = proj['readme']
        if isinstance(readme, str):
            lc.referenced_files.append(readme)
            desc_content, mimetype = description_from_file(readme, path.parent)

        elif isinstance(readme, dict):
            unrec_keys = set(readme.keys()) - {'text', 'file', 'content-type'}
            if unrec_keys:
                raise ConfigError(
                    "Unrecognised keys in [project.readme]: {}".format(unrec_keys)
                )
            if 'content-type' in readme:
                mimetype = readme['content-type']
                mtype_base = mimetype.split(';')[0].strip()  # e.g. text/x-rst
                if mtype_base not in readme_ext_to_content_type.values():
                    raise ConfigError(
                        "Unrecognised readme content-type: {!r}".format(mtype_base)
                    )
                # TODO: validate content-type parameters (charset, md variant)?
            else:
                raise ConfigError(
                    "content-type field required in [project.readme] table"
                )
            if 'file' in readme:
                if 'text' in readme:
                    raise ConfigError(
                        "[project.readme] should specify file or text, not both"
                    )
                lc.referenced_files.append(readme['file'])
                desc_content, _ = description_from_file(
                    readme['file'], path.parent, guess_mimetype=False
                )
            elif 'text' in readme:
                desc_content = readme['text']
            else:
                raise ConfigError(
                    "file or text field required in [project.readme] table"
                )
        else:
            raise ConfigError(
                "project.readme should be a string or a table"
            )

        md_dict['description'] = desc_content
        md_dict['description_content_type'] = mimetype

    if 'requires-python' in proj:
        md_dict['requires_python'] = proj['requires-python']

    license_files = set()
    if 'license' in proj:
        _check_types(proj, 'license', (str, dict))
        if isinstance(proj['license'], str):
            licence_expr = proj['license']
            md_dict['license_expression'] = normalise_compound_license_expr(licence_expr)
        else:
            license_tbl = proj['license']
            unrec_keys = set(license_tbl.keys()) - {'text', 'file'}
            if unrec_keys:
                raise ConfigError(
                    "Unrecognised keys in [project.license]: {}".format(unrec_keys)
                )

            # The 'License' field in packaging metadata is a brief description of
            # a license, not the full text or a file path.
            if 'file' in license_tbl:
                if 'text' in license_tbl:
                    raise ConfigError(
                        "[project.license] should specify file or text, not both"
                    )
                license_f = osp.normpath(license_tbl['file'])
                if isabs_ish(license_f):
                    raise ConfigError(
                        f"License file path ({license_tbl['file']}) cannot be an absolute path"
                    )
                if license_f.startswith('..' + os.sep):
                    raise ConfigError(
                        f"License file path ({license_tbl['file']}) cannot contain '..'"
                    )
                license_p = path.parent / license_f
                if not license_p.is_file():
                    raise ConfigError(f"License file {license_tbl['file']} does not exist")
                license_f = license_p.relative_to(path.parent).as_posix()
                license_files.add(license_f)
            elif 'text' in license_tbl:
                pass
            else:
                raise ConfigError(
                    "file or text field required in [project.license] table"
                )

    if 'license-files' in proj:
        _check_type(proj, 'license-files', list)
        globs = proj['license-files']
        license_files = _license_files_from_globs(path.parent, globs)
        if isinstance(proj.get('license'), dict):
            raise ConfigError(
                "license-files cannot be used with a license table, "
                "use 'project.license' with a license expression instead"
            )
    else:
        license_files.update(
            _license_files_from_globs(
                path.parent, default_license_files_globs, warn_no_files=False
            )
        )
    license_files_sorted = sorted(license_files)
    lc.referenced_files.extend(license_files_sorted)
    md_dict['license_files'] = license_files_sorted

    if 'authors' in proj:
        _check_type(proj, 'authors', list)
        md_dict.update(pep621_people(proj['authors']))

    if 'maintainers' in proj:
        _check_type(proj, 'maintainers', list)
        md_dict.update(pep621_people(proj['maintainers'], group_name='maintainer'))

    if 'keywords' in proj:
        _check_list_of_str(proj, 'keywords')
        md_dict['keywords'] = ",".join(proj['keywords'])

    if 'classifiers' in proj:
        _check_list_of_str(proj, 'classifiers')
        classifiers = proj['classifiers']
        license_expr = md_dict.get('license_expression', None)
        if license_expr:
            for cl in classifiers:
                if not cl.startswith('License :: '):
                    continue
                raise ConfigError(
                    "License classifiers are deprecated in favor of the license expression. "
                    "Remove the '{}' classifier".format(cl)
                )
        md_dict['classifiers'] = proj['classifiers']

    if 'urls' in proj:
        _check_type(proj, 'urls', dict)
        project_urls = md_dict['project_urls'] = []
        for label, url in sorted(proj['urls'].items()):
            project_urls.append("{}, {}".format(label, url))

    if 'entry-points' in proj:
        _check_type(proj, 'entry-points', dict)
        for grp in proj['entry-points'].values():
            if not isinstance(grp, dict):
                raise ConfigError(
                    "projects.entry-points should only contain sub-tables"
                )
            if not all(isinstance(k, str) for k in grp.values()):
                raise ConfigError(
                    "[projects.entry-points.*] tables should have string values"
                )
        if set(proj['entry-points'].keys()) & {'console_scripts', 'gui_scripts'}:
            raise ConfigError(
                "Scripts should be specified in [project.scripts] or "
                "[project.gui-scripts], not under [project.entry-points]"
            )
        lc.entrypoints = proj['entry-points']

    if 'scripts' in proj:
        _check_type(proj, 'scripts', dict)
        if not all(isinstance(k, str) for k in proj['scripts'].values()):
            raise ConfigError(
                "[projects.scripts] table should have string values"
            )
        lc.entrypoints['console_scripts'] = proj['scripts']

    if 'gui-scripts' in proj:
        _check_type(proj, 'gui-scripts', dict)
        if not all(isinstance(k, str) for k in proj['gui-scripts'].values()):
            raise ConfigError(
                "[projects.gui-scripts] table should have string values"
            )
        lc.entrypoints['gui_scripts'] = proj['gui-scripts']

    if 'dependencies' in proj:
        _check_list_of_str(proj, 'dependencies')
        reqs_noextra = proj['dependencies']
    else:
        reqs_noextra = []

    if 'optional-dependencies' in proj:
        _check_type(proj, 'optional-dependencies', dict)
        optdeps = proj['optional-dependencies']
        if not all(isinstance(e, list) for e in optdeps.values()):
            raise ConfigError(
                'Expected a dict of lists in optional-dependencies field'
            )
        extra_names_by_normed = {}
        for e, reqs in optdeps.items():
            if not all(isinstance(a, str) for a in reqs):
                raise ConfigError(
                    'Expected a string list for optional-dependencies ({})'.format(e)
                )
            if not name_is_valid(e):
                raise ConfigError(
                    f'optional-dependencies group name {e!r} is not valid'
                )
            enorm = normalise_core_metadata_name(e)
            extra_names_by_normed.setdefault(enorm, set()).add(e)
            lc.reqs_by_extra[enorm] = reqs

        clashing_extra_names = [
            g for g in extra_names_by_normed.values() if len(g) > 1
        ]
        if clashing_extra_names:
            fmted = ['/'.join(sorted(g)) for g in clashing_extra_names]
            raise ConfigError(
                f"optional-dependencies group names clash: {'; '.join(fmted)}"
            )

        md_dict['provides_extra'] = sorted(lc.reqs_by_extra.keys())

    md_dict['requires_dist'] = \
        reqs_noextra + list(_expand_requires_extra(lc.reqs_by_extra))

    # For internal use, record the main requirements as a '.none' extra.
    if reqs_noextra:
        lc.reqs_by_extra['.none'] = reqs_noextra

    if 'dynamic' in proj:
        _check_list_of_str(proj, 'dynamic')
        dynamic = set(proj['dynamic'])
        unrec_dynamic = dynamic - {'version', 'description'}
        if unrec_dynamic:
            raise ConfigError(
                "flit only supports dynamic metadata for 'version' & 'description'"
            )
        if dynamic.intersection(proj):
            raise ConfigError(
                "keys listed in project.dynamic must not be in [project] table"
            )
        lc.dynamic_metadata = dynamic

    if ('version' not in proj) and ('version' not in lc.dynamic_metadata):
        raise ConfigError(
            "version must be specified under [project] or listed as a dynamic field"
        )
    if ('description' not in proj) and ('description' not in lc.dynamic_metadata):
        raise ConfigError(
            "description must be specified under [project] or listed as a dynamic field"
        )

    return lc


def name_is_valid(name) -> bool:
    return bool(re.match(
        r"^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$", name, re.IGNORECASE
    ))


def pep621_people(people, group_name='author') -> dict:
    """Convert authors/maintainers from PEP 621 to core metadata fields"""
    names, emails = [], []
    for person in people:
        if not isinstance(person, dict):
            raise ConfigError("{} info must be list of dicts".format(group_name))
        unrec_keys = set(person.keys()) - {'name', 'email'}
        if unrec_keys:
            raise ConfigError(
                "Unrecognised keys in {} info: {}".format(group_name, unrec_keys)
            )
        if 'email' in person:
            email = person['email']
            if 'name' in person:
                email = str(Address(person['name'], addr_spec=email))
            emails.append(email)
        elif 'name' in person:
            names.append(person['name'])

    res = {}
    if names:
        res[group_name] = ", ".join(names)
    if emails:
        res[group_name + '_email'] = ", ".join(emails)
    return res


def isabs_ish(path):
    """Like os.path.isabs(), but Windows paths from a drive root count as absolute

    isabs() worked this way up to Python 3.12 (inclusive), and where we reject
    absolute paths, we also want to reject these odd halfway paths.
    """
    return os.path.isabs(path) or path.startswith(('/', '\\'))


def normalise_compound_license_expr(s: str) -> str:
    """Validate and normalise a compund SPDX license expression.

    Per the specification, licence expression operators (AND, OR and WITH)
    are matched case-sensitively. The WITH operator is not currently supported.

    Spec: https://spdx.github.io/spdx-spec/v2.2.2/SPDX-license-expressions/
    """
    invalid_msg = "'{s}' is not a valid SPDX license expression: {reason}"
    if not s or s.isspace():
        raise ConfigError(f"The SPDX license expression must not be empty")

    stack = 0
    parts = []
    try:
        for part in filter(None, re.split(r' +|([()])', s)):
            if part.upper() == 'WITH':
                # provide a sensible error message for the WITH operator
                raise ConfigError(f"The SPDX 'WITH' operator is not yet supported!")
            elif part in {'AND', 'OR'}:
                if not parts or parts[-1] in {' AND ', ' OR ', ' WITH ', '('}:
                    reason = f"a license ID is missing before '{part}'"
                    raise ConfigError(invalid_msg.format(s=s, reason=reason))
                parts.append(f' {part} ')
            elif part.lower() in {'and', 'or', 'with'}:
                # provide a sensible error message for non-uppercase operators
                reason = f"operators must be uppercase, not '{part}'"
                raise ConfigError(invalid_msg.format(s=s, reason=reason))
            elif part == '(':
                if parts and parts[-1] not in {' AND ', ' OR ', '('}:
                    reason = f"'(' must follow either AND, OR, or '('"
                    raise ConfigError(invalid_msg.format(s=s, reason=reason))
                stack += 1
                parts.append(part)
            elif part == ')':
                if not parts or parts[-1] in {' AND ', ' OR ', ' WITH ', '('}:
                    reason = f"a license ID is missing before '{part}'"
                    raise ConfigError(invalid_msg.format(s=s, reason=reason))
                stack -= 1
                if stack < 0:
                    reason = 'unbalanced brackets'
                    raise ConfigError(invalid_msg.format(s=s, reason=reason))
                parts.append(part)
            else:
                if parts and parts[-1] not in {' AND ', ' OR ', '('}:
                    reason = f"a license ID must follow either AND, OR, or '('"
                    raise ConfigError(invalid_msg.format(s=s, reason=reason))
                simple_expr = normalise_simple_license_expr(part)
                parts.append(simple_expr)

        if stack != 0:
            reason = 'unbalanced brackets'
            raise ConfigError(invalid_msg.format(s=s, reason=reason))
        if parts[-1] in {' AND ', ' OR ', ' WITH '}:
            last_part = parts[-1].strip()
            reason = f"a license ID or expression should follow '{last_part}'"
            raise ConfigError(invalid_msg.format(s=s, reason=reason))
    except ConfigError:
        if os.environ.get('FLIT_ALLOW_INVALID'):
            log.warning(f"Invalid license ID {s!r} allowed by FLIT_ALLOW_INVALID")
            return s
        raise

    return ''.join(parts)


def normalise_simple_license_expr(s: str) -> str:
    """Normalise a simple SPDX license expression.

    https://spdx.github.io/spdx-spec/v2.2.2/SPDX-license-expressions/#d3-simple-license-expressions
    """
    ls = s.lower()
    if ls.startswith('licenseref-'):
        ref = s[11:]
        if re.fullmatch(r'[a-zA-Z0-9\-.]+', ref):
            # Normalise case of LicenseRef, leave the rest alone
            return f"LicenseRef-{ref}"
        raise ConfigError(
            "LicenseRef- license expression can only contain ASCII letters "
            "& digits, - and ."
        )

    or_later = ls.endswith('+')
    if or_later:
        ls = ls[:-1]

    try:
        normalised_id = licenses[ls]['id']
    except KeyError:
        raise ConfigError(f"{s!r} is not a recognised SPDX license ID")

    if or_later:
        return f'{normalised_id}+'
    return normalised_id



"""Core wheel installation logic."""

import posixpath
from io import BytesIO
from typing import Dict, Tuple, cast

from installer.destinations import WheelDestination
from installer.exceptions import InvalidWheelSource
from installer.records import RecordEntry
from installer.sources import WheelSource
from installer.utils import SCHEME_NAMES, Scheme, parse_entrypoints, parse_metadata_file

__all__ = ["install"]


def _process_WHEEL_file(source: WheelSource) -> Scheme:
    """Process the WHEEL file, from ``source``.

    Returns the scheme that the archive root should go in.
    """
    stream = source.read_dist_info("WHEEL")
    metadata = parse_metadata_file(stream)

    # Ensure compatibility with this wheel version.
    if not (metadata["Wheel-Version"] and metadata["Wheel-Version"].startswith("1.")):
        message = "Incompatible Wheel-Version {}, only support version 1.x wheels."
        raise InvalidWheelSource(source, message.format(metadata["Wheel-Version"]))

    # Determine where archive root should go.
    if metadata["Root-Is-Purelib"] == "true":
        return cast(Scheme, "purelib")
    else:
        return cast(Scheme, "platlib")


def _determine_scheme(
    path: str, source: WheelSource, root_scheme: Scheme
) -> Tuple[Scheme, str]:
    """Determine which scheme to place given path in, from source."""
    data_dir = source.data_dir

    # If it's in not `{distribution}-{version}.data`, then it's in root_scheme.
    if posixpath.commonprefix([data_dir, path]) != data_dir:
        return root_scheme, path

    # Figure out which scheme this goes to.
    parts = []
    scheme_name = None
    left = path
    while True:
        left, right = posixpath.split(left)
        parts.append(right)
        if left == source.data_dir:
            scheme_name = right
            break

    if scheme_name not in SCHEME_NAMES:
        msg_fmt = "{path} is not contained in a valid .data subdirectory."
        raise InvalidWheelSource(source, msg_fmt.format(path=path))

    return cast(Scheme, scheme_name), posixpath.join(*reversed(parts[:-1]))


def install(
    source: WheelSource,
    destination: WheelDestination,
    additional_metadata: Dict[str, bytes],
) -> None:
    """Install wheel described by ``source`` into ``destination``.

    :param source: wheel to install.
    :param destination: where to write the wheel.
    :param additional_metadata: additional metadata files to generate, usually
                                generated by the caller.

    """
    root_scheme = _process_WHEEL_file(source)

    # RECORD handling
    record_file_path = posixpath.join(source.dist_info_dir, "RECORD")
    written_records = []

    # Write the entry_points based scripts.
    if "entry_points.txt" in source.dist_info_filenames:
        entrypoints_text = source.read_dist_info("entry_points.txt")
        for name, module, attr, section in parse_entrypoints(entrypoints_text):
            record = destination.write_script(
                name=name,
                module=module,
                attr=attr,
                section=section,
            )
            written_records.append((Scheme("scripts"), record))

    # Write all the files from the wheel.
    for record_elements, stream, is_executable in source.get_contents():
        source_record = RecordEntry.from_elements(*record_elements)
        path = source_record.path
        # Skip the RECORD, which is written at the end, based on this info.
        if path == record_file_path:
            continue

        # Figure out where to write this file.
        scheme, destination_path = _determine_scheme(
            path=path,
            source=source,
            root_scheme=root_scheme,
        )
        record = destination.write_file(
            scheme=scheme,
            path=destination_path,
            stream=stream,
            is_executable=is_executable,
        )
        written_records.append((scheme, record))

    # Write all the installation-specific metadata
    for filename, contents in additional_metadata.items():
        path = posixpath.join(source.dist_info_dir, filename)

        with BytesIO(contents) as other_stream:
            record = destination.write_file(
                scheme=root_scheme,
                path=path,
                stream=other_stream,
                is_executable=False,
            )
        written_records.append((root_scheme, record))

    written_records.append((root_scheme, RecordEntry(record_file_path, None, None)))
    destination.finalize_installation(
        scheme=root_scheme,
        record_file_path=record_file_path,
        records=written_records,
    )



# test_stackDicts.py -- Portage Unit Testing Functionality
# Copyright 2006 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.util import stack_dicts


class StackDictsTestCase(TestCase):
    def testStackDictsPass(self):
        tests = [
            ([{"a": "b"}, {"b": "c"}], {"a": "b", "b": "c"}, False, [], False),
            ([{"a": "b"}, {"a": "c"}], {"a": "b c"}, True, [], False),
            ([{"a": "b"}, {"a": "c"}], {"a": "b c"}, False, ["a"], False),
            ([{"a": "b"}, None], {"a": "b"}, False, [], True),
            ([None], {}, False, [], False),
            ([None, {}], {}, False, [], True),
        ]
        for test in tests:
            result = stack_dicts(test[0], test[2], test[3], test[4])
            self.assertEqual(result, test[1])

    def testStackDictsFail(self):
        tests = [
            ([None, {}], None, False, [], True),
            ([{"a": "b"}, {"a": "c"}], {"a": "b c"}, False, [], False),
        ]
        for test in tests:
            result = stack_dicts(test[0], test[2], test[3], test[4])
            self.assertNotEqual(result, test[1])
